##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
pykiso - extensible framework for (embedded) integration testing.
*****************************************************************

:module: pykiso

:synopsis: ``pykiso`` is an extensible framework for (embedded) integration testing.

.. currentmodule:: pykiso

"""

__version__ = "0.16.0"


from . import auxiliary, cli, config_parser, connector, message, types
from .auxiliary import AuxiliaryCommon
from .connector import CChannel, Flasher
from .interfaces.mp_auxiliary import MpAuxiliaryInterface
from .interfaces.simple_auxiliary import SimpleAuxiliaryInterface
from .interfaces.thread_auxiliary import AuxiliaryInterface
from .message import Message
from .test_coordinator import test_case, test_message_handler, test_suite
from .test_coordinator.test_case import (
    BasicTest,
    define_test_parameters,
    retry_test_case,
)
from .test_coordinator.test_suite import (
    BasicTestSuiteSetup,
    BasicTestSuiteTeardown,
)



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Auxiliary common Interface Definition
*************************************

:module: auxiliary

:synopsis: base auxiliary interface

.. currentmodule:: pykiso

"""
from __future__ import annotations

import abc
import logging
import queue
import time
from typing import Any

from .test_setup.config_registry import ConfigRegistry
from .types import MsgType

log = logging.getLogger(__name__)


class AuxiliaryCommon(metaclass=abc.ABCMeta):
    """Class use to encapsulate all common methods/attributes for both
    multiprocessing and thread auxiliary interface.
    """

    def __init__(self) -> None:
        """Auxiliary common attributes initialization."""
        self.name = None
        self.queue_in = None
        self.lock = None
        self.queue_out = None
        self.is_instance = False
        self.stop_event = None
        self._aux_copy = None

    def __repr__(self) -> str:
        name = self.name
        repr_ = super().__repr__()
        if name:
            repr_ = repr_[:1] + f"{name} is " + repr_[1:]
        return repr_

    def lock_it(self, timeout_in_s: float) -> bool:
        """Lock to ensure exclusivity.

        :param timeout_in_s: How many second you want to wait for the lock

        :return: True - Lock done / False - Lock failed
        """
        return self.lock.acquire(timeout=timeout_in_s)

    def unlock_it(self) -> None:
        """Unlock exclusivity"""
        self.lock.release()

    def create_copy(self, *args, **config: dict) -> AuxiliaryCommon:
        """Create a copy of the actual auxiliary instance with the
        new desired configuration.

        .. note:: only named arguments have to be used

        .. warning:: the call of create_copy will automatically suspend
            the current auxiliary until the it copy is destroyed

        :param config: new desired auxiliary configuration

        :return: a brand new auxiliary instance

        :raises Exception: if positional parameters is given or unknown
            named parameters are given
        """
        # only named parameters are allowed
        if args:
            raise Exception("Only use named parameters when invoking create_copy")
        # if a copy already exist return the actual one
        if self._aux_copy is not None:
            log.warning(
                f"A copy of {self} already exists, destroy it before creating a new one"
            )
            return self._aux_copy
        # get modified parameters based on the yaml one
        base_conf = modified_conf = ConfigRegistry.get_aux_config(self.name)
        modified_params = list(set(config) & set(base_conf))
        added_params = list(set(config) - set(base_conf))
        if self.is_alive():
            self.suspend()
            # add this timeout if cc_proxy connectors are used (avoid
            # possible ConnectionRefusedError)
            time.sleep(1.100)
        # apply new configuration parameters values based on yaml config
        for name, val in base_conf.items():
            if name in modified_params:
                modified_conf[name] = config[name]
        # add new parameters values (not explicitly mention in yaml)
        for name in added_params:
            modified_conf[name] = config[name]
        try:
            # create a brand new auxiliary instance
            self._aux_copy = self.__class__(**modified_conf)
        except TypeError:
            raise Exception(
                f"Unknown parameter(s) given to {self.name} see {added_params}"
            )
        self._aux_copy.name = self.name
        auto_start = getattr(self._aux_copy, "auto_start", None)
        if auto_start:
            self._aux_copy.start()
            self._aux_copy.create_instance()
        return self._aux_copy

    def destroy_copy(self) -> None:
        """Stop the current auxiliary copy and resume the original.

        .. warning:: stop the copy auxiliary will automatically start
            the base/original one
        """
        if self._aux_copy is not None:
            # delete and stop the thread properly
            self._aux_copy.delete_instance()
            self._aux_copy.stop()
            self._aux_copy.join()
            self._aux_copy = None
        if not self.is_alive():
            self.start()
        if not self.is_instance:
            self.resume()

    def run_command(
        self,
        cmd_message: MsgType,
        cmd_data: Any = None,
        blocking: bool = True,
        timeout_in_s: int = 0,
    ) -> bool:
        """Send a test request.

        :param cmd_message: command request to the auxiliary
        :param cmd_data: data you would like to populate the command with
        :param blocking: If you want the command request to be blocking or not
        :param timeout_in_s: Number of time (in s) you want to wait for an answer

        :return: True - Successfully sent / False - Failed by sending / None
        """
        return_code = False

        log.debug(f"sending command '{cmd_message}' in {self}")
        if cmd_data:
            log.debug(f"command payload data: {repr(cmd_data)}")

        if self.lock.acquire():
            # Trigger the internal requests
            self.queue_in.put(("command", cmd_message, cmd_data))
            log.debug(f"sent command '{cmd_message}' in {self}")
            # Wait until the test request was received
            try:
                log.debug(f"waiting for reply to command '{cmd_message}' in {self}")
                return_code = self.queue_out.get(blocking, timeout_in_s)
                log.debug(
                    f"reply to command '{cmd_message}' received: '{return_code}' in {self}"
                )
            except queue.Empty:
                log.debug("no reply received within time")
            # Release the above lock
            self.lock.release()
            # Return the ack_report if exists
        return return_code

    def abort_command(self, blocking: bool = True, timeout_in_s: float = 25) -> bool:
        """Force test to abort.

        :param blocking: If you want the command request to be blocking or not
        :param timeout_in_s: Number of time (in s) you want to wait
            for an answer

        :return: True - Abort was a success / False - if not
        """
        return_code = False

        if self.lock.acquire():
            # Trigger the internal requests
            self.queue_in.put("abort")
            # Wait until the test request was received
            try:
                return_code = self.queue_out.get(blocking, timeout_in_s)
            except queue.Empty:
                log.info("no reply received within time")
            # Release the above lock
            self.lock.release()
            # Return the ack_report if exists
        return return_code

    def wait_and_get_report(
        self, blocking: bool = False, timeout_in_s: int = 0
    ) -> MsgType:
        """Wait for the report of the previous sent test request.

        :param blocking: True: wait for timeout to expire, False: return immediately
        :param timeout_in_s: if blocking, wait the defined time in seconds

        :return: a message.Message() - Message received / None - nothing received
        """
        try:
            return self.queue_out.get(blocking, timeout_in_s)
        except queue.Empty:
            return None

    def stop(self) -> None:
        """Force the thread to stop itself."""
        if self._aux_copy is not None:
            # if an auxiliary copy remains just stop it before the
            # original one
            self._aux_copy.delete_instance()
            self._aux_copy.stop()
            self._aux_copy.join()
            self._aux_copy = None
        self.stop_event.set()

    def resume(self) -> None:
        """Resume current auxiliary's run, by running the
        create_instance method in the background.

        .. warning:: due to the usage of create_instance if an issue
            occurred the exception AuxiliaryCreationError is raised.
        """
        if not self.stop_event.is_set() and not self.is_instance:
            self.create_instance()
        else:
            log.warning(f"Auxiliary '{self}' is already running")

    def suspend(self) -> None:
        """Supend current auxiliary's run."""
        if self.is_instance:
            self.delete_instance()
        else:
            log.warning(f"Auxiliary '{self}' is already stopped")

    @abc.abstractmethod
    def create_instance(self) -> bool:
        """Handle auxiliary creation."""
        pass

    @abc.abstractmethod
    def delete_instance(self) -> bool:
        """Handle auxiliary deletion."""
        pass

    @abc.abstractmethod
    def run(self) -> None:
        """Run function of the auxiliary."""
        pass



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Multiprocessing Proxy Channel
*****************************

:module: cc_mp_proxy

:synopsis: concrete implementation of a multiprocessing proxy channel

CCProxy channel was created, in order to enable the connection of
multiple auxiliaries on one and only one CChannel. This CChannel
has to be used with a so called proxy auxiliary.

.. currentmodule:: cc_mp_proxy

"""

import logging
from multiprocessing import Queue

from pykiso.lib.connectors.cc_proxy import CCProxy

log = logging.getLogger(__name__)


class CCMpProxy(CCProxy):
    """Multiprocessing Proxy CChannel for multi auxiliary usage."""

    def __init__(self, **kwargs):
        """Initialize attributes."""

        super().__init__(**kwargs)
        # instantiate directly both queue_in and queue_out
        self.queue_in = Queue()
        self.queue_out = Queue()
        self.timeout = 1

    def _cc_open(self) -> None:
        """Open proxy channel.

        Due to usage of multiprocessing the queue_in and queue_out state
        doesn't have to change in order to ensure that ProxyAuxiliary
        works even if suspend or resume is called.
        """
        log.debug("Open proxy channel")

    def _cc_close(self) -> None:
        """Close proxy channel.

        Due to usage of multiprocessing the queue_in and queue_out state
        doesn't have to change in order to ensure that ProxyAuxiliary
        works even if suspend or resume is called.
        """
        log.debug("Close proxy channel")



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Test Suite
**********

:module: test_suite

:synopsis: Create a generic test-suite based on the connected modules

.. currentmodule:: generic_test_suite


"""

import logging
import unittest
from collections.abc import Iterable
from typing import Callable, List, Union

from .. import message
from ..interfaces.thread_auxiliary import AuxiliaryInterface
from .test_message_handler import test_app_interaction

__all__ = [
    "BaseTestSuite",
    "BasicTestSuiteSetup",
    "BasicTestSuiteTeardown",
    "BasicTestSuite",
]

log = logging.getLogger(__name__)


class BaseTestSuite(unittest.TestCase):

    response_timeout = 10

    def __init__(
        self,
        test_suite_id: int,
        test_case_id: int,
        aux_list: Union[List[AuxiliaryInterface], None],
        setup_timeout: Union[int, None],
        run_timeout: Union[int, None],
        teardown_timeout: Union[int, None],
        test_ids: Union[dict, None],
        variant: Union[list, None],
        args: tuple,
        kwargs: dict,
    ):
        """Initialize generic test-case.

        :param test_suite_id: test suite identification number
        :param test_case_id: test case identification number
        :param aux_list: list of used auxiliaries
        :param setup_timeout: maximum time (in seconds) used to wait
            for a report during setup execution
        :param run_timeout: maximum time (in seconds) used to wait for
            a report during test_run execution
        :param teardown_timeout: the maximum time (in seconds) used to
            wait for a report during teardown execution
        :param test_ids: jama references to get the coverage
            eg: {"Component1": ["Req1", "Req2"], "Component2 ["Req3"]}
        :param variant: string that allows the user to execute a subset of tests
        """
        # Initialize base class
        super().__init__(*args, **kwargs)
        # Save test information
        self.test_auxiliary_list = aux_list or []
        self.test_suite_id = test_suite_id
        self.test_case_id = test_case_id
        self.setup_timeout = setup_timeout or BaseTestSuite.response_timeout
        self.run_timeout = run_timeout or BaseTestSuite.response_timeout
        self.teardown_timeout = teardown_timeout or BaseTestSuite.response_timeout
        self.test_ids = test_ids
        self.variant = variant

    def cleanup_and_skip(self, aux: AuxiliaryInterface, info_to_print: str):
        """Cleanup auxiliary and log reasons.

        :param aux: corresponding auxiliary to abort
        :param info_to_print: A message you want to print while cleaning up the test
        """
        # Log error message
        log.critical(info_to_print)

        # Send aborts to corresponding auxiliary
        if aux.abort_command() is not True:
            log.critical(f"Error occurred during abort command on auxiliary {aux}")

        self.fail(info_to_print)


class BasicTestSuiteSetup(BaseTestSuite):
    """Inherit from unittest testCase and represent setup fixture."""

    def __init__(
        self,
        test_suite_id: int,
        test_case_id: int,
        aux_list: Union[List[AuxiliaryInterface], None],
        setup_timeout: Union[int, None],
        run_timeout: Union[int, None],
        teardown_timeout: Union[int, None],
        test_ids: Union[dict, None],
        variant: Union[list, None],
        args: tuple,
        kwargs: dict,
    ):
        """Initialize generic test-case.

        :param test_suite_id: test suite identification number
        :param test_case_id: test case identification number
        :param aux_list: list of used auxiliaries
        :param setup_timeout: maximum time (in seconds) used to wait
            for a report during setup execution
        :param run_timeout: maximum time (in seconds) used to wait for
            a report during test_run execution
        :param teardown_timeout: the maximum time (in seconds) used to
            wait for a report during teardown execution
        :param test_ids: jama references to get the coverage
            eg: {"Component1": ["Req1", "Req2"], "Component2": ["Req3"]}
        :param variant: string that allows the user to execute a subset of tests
        """
        # Initialize base class
        super().__init__(
            test_suite_id,
            test_case_id,
            aux_list,
            setup_timeout,
            run_timeout,
            teardown_timeout,
            test_ids,
            variant,
            args,
            kwargs,
        )

    @test_app_interaction(
        message_type=message.MessageCommandType.TEST_SUITE_SETUP, timeout_cmd=5
    )
    def test_suite_setUp(self):
        """Test method for constructing the actual test suite."""
        pass


class BasicTestSuiteTeardown(BaseTestSuite):
    """Inherit from unittest testCase and represent teardown fixture."""

    def __init__(
        self,
        test_suite_id: int,
        test_case_id: int,
        aux_list: Union[List[AuxiliaryInterface], None],
        setup_timeout: Union[int, None],
        run_timeout: Union[int, None],
        teardown_timeout: Union[int, None],
        test_ids: Union[dict, None],
        variant: Union[list, None],
        args: tuple,
        kwargs: dict,
    ):
        """Initialize generic test-case.

        :param test_suite_id: test suite identification number
        :param test_case_id: test case identification number
        :param aux_list: list of used auxiliaries
        :param setup_timeout: maximum time (in seconds) used to wait
            for a report during setup execution
        :param run_timeout: maximum time (in seconds) used to wait for
            a report during test_run execution
        :param teardown_timeout: the maximum time (in seconds) used to
            wait for a report during teardown execution
        :param test_ids: jama references to get the coverage
            eg: {"Component1": ["Req1", "Req2"], "Component2": ["Req3"]}
        :param variant: string that allows the user to execute a subset of tests
        """
        # Initialize base class
        super().__init__(
            test_suite_id,
            test_case_id,
            aux_list,
            setup_timeout,
            run_timeout,
            teardown_timeout,
            test_ids,
            variant,
            args,
            kwargs,
        )

    @test_app_interaction(
        message_type=message.MessageCommandType.TEST_SUITE_TEARDOWN, timeout_cmd=5
    )
    def test_suite_tearDown(self):
        """Test method for deconstructing the actual test suite after testing it."""
        pass


class BasicTestSuite(unittest.TestSuite):
    """Inherit from the unittest framework test-suite but build it for our integration tests."""

    def __init__(
        self,
        modules_to_add_dir: str,
        test_filter_pattern: str,
        test_suite_id: int,
        args: tuple,
        kwargs: dict,
    ):
        """Initialize our custom unittest-test-suite.

        .. note::
            1. Will Load from the given path the integration test modules under test
            2. Sort the given test case list by test suite/case id
            3. Place Test suite setup and teardown respectively at top and bottom of test case list
            4. Add sorted test case list to test suite
        """
        # Mother class initialization
        super().__init__(*args, **kwargs)

        # load test from the specified folder
        loader = unittest.TestLoader()
        found_modules = loader.discover(modules_to_add_dir, pattern=test_filter_pattern)

        # sort the test case list by ascendant using test suite and test case id
        test_case_list = sorted(flatten(found_modules), key=tc_sort_key)

        # add sorted test case list to test suite
        self.addTests(test_case_list)


def tc_sort_key(tc):
    """Sort-key for testcases.

    will sort by test-suite/test-case, but the setup will always be first,
    the teardown will always be last.

    :param tc: a BaseTestSuite/TestCase to rank

    :return: key for :py:func:`sorted`

    :raise: any exception that occurs during test loading
    """
    try:
        fix_ind = 0
        if isinstance(tc, BasicTestSuiteSetup):
            fix_ind = -1
        elif isinstance(tc, BasicTestSuiteTeardown):
            fix_ind = 1
        elif isinstance(tc, unittest.loader._FailedTest):
            raise tc._exception
        return (fix_ind, tc.test_suite_id, tc.test_case_id)
    except BaseException:
        log.exception("Issue detected during test suite initialization!")


def flatten(it):
    """Flatten all level of nesting.

    :param it: nested iterable

    :return: first not nested items
    """
    for x in it:
        if isinstance(x, Iterable) and not isinstance(x, (str, bytes)):
            yield from flatten(x)
        else:
            yield x



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Handle common communication with device under test
**************************************************

By default, the integration test framework handles internal messaging
and control flow using a message format defined in
:py:class:`pykiso.Message`. :py:mod:`pykiso.test_message_handler`
defines the default messaging protocol from a behavioral point of view.

The general procedure is described in handle_basic_interaction context
manager, but specific _MsgHandler_ classes are  provided with
:py:class:`TestCaseMsgHandler` and :py:class:`TestSuiteMsgHandler` to
provide shorthands for the specialised communication from
:py:class:`pykiso.test_case.BasicTest` and
:py:class:`pykiso.test_suite.BasicTestSuite`.


:module: test_message_handler

:synopsis: default communication between TestManagement and DUT.

.. currentmodule:: test_message_handler

"""
import collections
import functools
import logging
from contextlib import contextmanager
from typing import Callable, List

from pykiso import message

__all__ = [
    "report_analysis",
    "handle_basic_interaction",
    "test_app_interaction",
]

report_analysis = collections.namedtuple(
    "report_analysis",
    ["current_auxiliary", "report_message", "logging_method", "log_message"],
)

cmd_response = collections.namedtuple(
    "cmd_response", ["valid", "sent_command", "current_auxiliary"]
)

wait_report = collections.namedtuple("wait_report", ["current_auxiliary", "message"])

log = logging.getLogger(__name__)


@contextmanager
def handle_basic_interaction(
    test_entity: Callable,
    cmd_sub_type: message.MessageCommandType,
    timeout_cmd: int,
    timeout_resp: int,
) -> List[report_analysis]:
    """Handle default communication mechanism between test manager and device under
    test as follow:

    .. code-block:: none

       TM |   COMMAND ---------->        | DUT
       TM |           <---------- ACK    | DUT
       TM |           <---------- LOG    | DUT
       TM |   ACK     ---------->        | DUT
       ...
       TM |           <---------- LOG    | DUT
       TM |   ACK     ---------->        | DUT
       TM |           <---------- REPORT | DUT
       TM |   ACK     ---------->        | DUT

    This behaviour is implemented here.

    Logs can be sent to TM while waiting for report.

    :param test_entity: test instance in use (BaseTestSuite, BasicTest,...)
    :param cmd_sub_type: message command sub-type (Test case run, setup,....)
    :param timeout_cmd: timeout in seconds for auxiliary run_command
    :param timeout_resp: timeout in seconds for auxiliary wait_and_get_report

    :return: tuple containing current auxiliary, reported message, logging method to use,
        and pre-defined log message.
    """
    responses = []
    # send command and check if DUT response is correctly received
    with Command.send(
        cmd_sub_type=cmd_sub_type, test_entity=test_entity, timeout_cmd=timeout_cmd
    ) as cmd_responses:

        for cmd_execution in cmd_responses:
            if not cmd_execution.valid:
                info_to_print = f"No response received from DUT for auxiliairy : {cmd_execution.current_auxiliary} command : {cmd_execution.sent_command}!"
                test_entity.cleanup_and_skip(
                    cmd_execution.current_auxiliary, info_to_print
                )

            # wait for DUT log and report and evaluate the response
            # while loop to wait for reports while getting the logs
            else:
                report_received = False
                while not report_received:
                    received_msg = cmd_execution.current_auxiliary.wait_and_get_report(
                        blocking=True, timeout_in_s=timeout_resp
                    )
                    if received_msg is None:
                        info_to_print = (
                            f"No report received from DUT for auxiliairy : {cmd_execution.current_auxiliary} command :{cmd_execution.sent_command}!",
                        )
                        test_entity.cleanup_and_skip(
                            cmd_execution.current_auxiliary, info_to_print
                        )
                        break
                    if received_msg.get_message_type() == message.MessageType.REPORT:
                        report_received = True
                    responses.append(
                        Command.evaluate_message(
                            cmd_execution.current_auxiliary, received_msg
                        )
                    )

    yield responses


class Command:
    """Encapsulate message command handling"""

    @classmethod
    @contextmanager
    def send(
        cls,
        test_entity: Callable,
        cmd_sub_type: message.MessageCommandType,
        timeout_cmd: int,
    ):
        """Used to send a specific message command to device under test for
        all current auxiliaries in use.

        :param test_entity: test instance in use (BaseTestSuite, BasicTest,...)
        :param cmd_sub_type: message command sub-type (Test case run, setup,....)
        :param timeout_cmd: timeout in second apply on auxiliary run_command

        :return: generator with namedtuple containing run_command verdict,
            command sent, and current auxiliary in use.
        """

        cmd = message.Message(
            msg_type=message.MessageType.COMMAND,
            sub_type=cmd_sub_type,
            test_suite=test_entity.test_suite_id,
            test_case=test_entity.test_case_id,
        )
        responses = []
        for aux in test_entity.test_auxiliary_list:
            if not aux.stop_event.is_set():
                _response = aux.run_command(
                    cmd, blocking=True, timeout_in_s=timeout_cmd
                )
                responses.append(cmd_response(_response, cmd, aux))
            else:
                log.fatal(f"Auxiliary {aux} is stopped")
                test_entity.fail(
                    f"Auxiliary {aux} is stopped: test suite id: {test_entity.test_suite_id}, test case id: {test_entity.test_case_id}"
                )
        yield responses

    @classmethod
    def evaluate_message(cls, aux, report_msg: message.Message) -> report_analysis:
        """Evaluate message type coming from DUT (COMMAND, LOG, REPORT...)

        :param aux: current auxiliary in use
        :param report_msg: message received from DUT

        :return: namedtuple containing current auxiliary, reported message, logging method to use,
            and pre-defined log message.
        """

        if report_msg.get_message_type() == message.MessageType.REPORT:
            return Report.evaluate_report(aux, report_msg)

        elif report_msg.get_message_type() == message.MessageType.LOG:
            log_msg = f"Logging message received from {aux.name},  : {report_msg}"
            return report_analysis(aux, report_msg, log.info, log_msg)
        else:
            log_msg = f"Message type unknown received from {aux.name},  : {report_msg}"
            return report_analysis(aux, report_msg, log.warning, log_msg)


class Report:
    """Manage Test App report."""

    @classmethod
    @contextmanager
    def wait(cls, auxiliaries, timeout: int):
        """Used to wait deveice under test's message command report.

        :param auxiliaries: auxiliar(y/ies) in use
        :param timeout: timeout in second apply on auxiliary wait_and_get_report

        :return: generator with namedtuple containing current auxiliary and current
            message from device under test(if timeout reach None otherwise Message instance)
        """

        for aux in auxiliaries:
            _report = aux.wait_and_get_report(blocking=True, timeout_in_s=timeout)
            yield wait_report(aux, _report)

    @classmethod
    def evaluate_report(cls, aux, report_msg: message.Message):
        """Evaluate message command report content (PASS, FAILED, NOT IMPLEMENTED)

        :param aux: current auxiliary in use
        :param report_msg: message received from DUT

        :return: namedtuple containing current auxiliary, reported message, logging method to use,
            and pre-defined log message.
        """
        log_fun = None
        log_msg = None
        if report_msg.get_message_sub_type() == message.MessageReportType.TEST_FAILED:
            log_fun = log.error
            log_msg = (
                f"Report with verdict FAILED received from {aux.name} : {report_msg}"
            )
        elif report_msg.get_message_sub_type() == message.MessageReportType.TEST_PASS:
            log_fun = log.info
            log_msg = (
                f"Report with verdict PASS received from {aux.name} : {report_msg}"
            )

        elif (
            report_msg.get_message_sub_type()
            == message.MessageReportType.TEST_NOT_IMPLEMENTED
        ):
            log_fun = log.info
            log_msg = (
                f"Report with verdict NOT IMPLEMENTED from {aux.name} : {report_msg}"
            )
        else:
            log_fun = log.warning
            log_msg = f"Report type unknown received from {aux.name} : {report_msg}"

        return report_analysis(aux, report_msg, log_fun, log_msg)


def test_app_interaction(
    message_type: message.MessageCommandType, timeout_cmd: int = 5
) -> Callable:
    """Handle test app basic interaction depending on the decorated
    method.

    :param message_type: message command sub-type (test case/suite
        run, setup, teardown....)
    :param timeout_cmd: timeout in seconds for auxiliary run_command

    :return: inner decorator function
    """

    def inner_interaction(func: Callable) -> Callable:
        """Inner decorator function.

        :param func: decorated method

        :return: decorator inner function
        """

        @functools.wraps(func)
        def handle_interaction(self, *args: tuple, **kwargs: dict) -> None:
            """Handle Test App communication mechanism during for all
            available fixtures(setup, run, teardown,...)

            :param args: positional arguments
            :param kwargs: named arguments

            :return: decorated method return (actual case None)
            """

            fixture = func.__name__.lower()

            if "setup" in fixture and "suite" in fixture:
                log.info(
                    f"--------------- SUITE SETUP: {self.test_suite_id} ---------------"
                )
                timeout_resp = self.setup_timeout
            elif "setup" in fixture:
                log.info(
                    f"--------------- SETUP: {self.test_suite_id}, {self.test_case_id} ---------------"
                )
                timeout_resp = self.setup_timeout
            elif "run" in fixture:
                log.info(
                    f"--------------- RUN: {self.test_suite_id}, {self.test_case_id} ---------------"
                )
                timeout_resp = self.run_timeout
            elif "teardown" in fixture and "suite" in fixture:
                log.info(
                    f"--------------- SUITE TEARDOWN: {self.test_suite_id} ---------------"
                )
                timeout_resp = self.teardown_timeout
            elif "teardown" in fixture:
                log.info(
                    f"--------------- TEARDOWN: {self.test_suite_id}, {self.test_case_id} ---------------"
                )
                timeout_resp = self.teardown_timeout

            # lock auxiliaries
            for aux in self.test_auxiliary_list:
                locked = aux.lock_it(1)
                if not locked:
                    self.cleanup_and_skip(aux, f"{aux} could not be locked!")

            with handle_basic_interaction(
                self, message_type, timeout_cmd, timeout_resp
            ) as report_infos:

                # Unlock all auxiliaries
                for aux in self.test_auxiliary_list:
                    aux.unlock_it()

                for aux, report_msg, log_level_func, log_msg in report_infos:
                    log_level_func(log_msg)

                    is_test_on_dut_implemented = (
                        report_msg.sub_type
                        != message.MessageReportType.TEST_NOT_IMPLEMENTED
                    )
                    is_report = (
                        report_msg.get_message_type() == message.MessageType.REPORT
                    )
                    if is_test_on_dut_implemented and is_report:
                        self.assertEqual(
                            report_msg.sub_type, message.MessageReportType.TEST_PASS
                        )

            return func(self, *args, **kwargs)

        return handle_interaction

    return inner_interaction



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
:module: exceptions

:synopsis: Define all custom exceptions raised by pykiso framework

.. currentmodule:: exceptions
"""


class PykisoError(Exception):
    """Pykiso specific exception used as basis for all others."""

    def __str__(self):
        return self.message


class AuxiliaryCreationError(PykisoError):
    """Raised when an auxiliary instance creation failed."""

    def __init__(self, aux_name: str) -> None:
        """Initialize attributes.

        :param aux_name: aux alias
        """
        self.message = f"Auxiliary {aux_name} failed at instance creation"
        super().__init__(self.message)



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Can Communication Channel SocketCAN
***********************************

:module: cc_socket_can

:synopsis: CChannel implementation for CAN(fd) using SocketCAN

.. currentmodule:: cc_socket_can

"""

import logging
import platform
import time
from pathlib import Path
from typing import Union

import can
import can.bus

from pykiso import CChannel, Message
from pykiso.lib.connectors.cc_socket_can.socketcan_to_trc import (
    SocketCan2Trc,
    can,
)

MessageType = Union[Message, bytes]

log = logging.getLogger(__name__)


def os_name() -> str:
    """Returns the system/OS name.

    :return: os name such as 'Linux', 'Darwin', 'Java', 'Windows'
    """
    return platform.system()


class CCSocketCan(CChannel):
    """CAN FD channel-adapter."""

    def __init__(
        self,
        channel: str = "vcan0",
        remote_id: int = None,
        is_fd: bool = True,
        enable_brs: bool = False,
        can_filters: list = None,
        is_extended_id: bool = False,
        receive_own_messages: bool = False,
        logging_activated: bool = False,
        log_path: str = None,
        log_name: str = None,
        **kwargs,
    ):
        """Initialize can channel settings.

        :param channel: the can interface name. (i.e. vcan0, can1, ..)
        :param remote_id: id used for transmission
        :param is_fd: should the Bus be initialized in CAN-FD mode
        :param enable_brs: sets the bitrate_switch flag to use higher transmission speed
        :param can_filters: iterable used to filter can id on reception
        :param is_extended_id: this flag controls the size of the arbitration_id field
        :param receive_own_messages: if set transmitted messages will be received
        :param logging_activated: boolean used to enable logfile creation
        :param log_path: trace directory path (absolute or relative)
        :param log_name: trace full name (without file extension)
        """
        super().__init__(**kwargs)
        self.channel = channel
        self.remote_id = remote_id
        self.is_fd = is_fd
        self.enable_brs = enable_brs
        self.can_filters = can_filters if can_filters else []
        self.is_extended_id = is_extended_id
        self.receive_own_messages = receive_own_messages
        self.logging_activated = logging_activated
        self.bus = None
        self.logger = None
        self.log_path = log_path
        self.log_name = log_name
        # Set a timeout to send the signal to the GIL to change thread.
        # In case of a multi-threading system, all tasks will be called one after the other.
        self.timeout = 1e-6

        if self.logging_activated:
            # Just avoid the case the given trace directory is None
            self.log_path = "" if self.log_name is None else self.log_name
            # if the given log path is not absolute add root path
            # (where pykiso is launched) otherwise take it as it is
            dir_path = (
                (Path() / self.log_path).resolve()
                if not Path(self.log_path).is_absolute()
                else Path(self.log_path)
            )
            # if no specific logging file name is given take the default one
            self.log_name = (
                time.strftime(f"%Y-%m-%d_%H-%M-%S_{self.log_name}.trc")
                if self.log_name is not None
                else time.strftime("%Y-%m-%d_%H-%M-%S_CanLog.trc")
            )
            # if path doesn't exists take root path (where pykiso is launched)
            self.log_path = (
                dir_path / self.log_name
                if dir_path.exists()
                else (Path() / self.log_name).resolve()
            )

        if self.enable_brs and not self.is_fd:
            log.warning(
                "Bitrate switch will have no effect because option is_fd is set to false."
            )

    def _cc_open(self) -> None:
        """Open a can bus channel, set filters for reception and activate"""

        if not os_name() == "Linux":
            raise OSError("socketCAN is only available under linux.")

        self.bus = can.interface.Bus(
            interface="socketcan",
            channel=self.channel,
            can_filters=self.can_filters,
            fd=self.is_fd,
            receive_own_messages=self.receive_own_messages,
        )

        if self.logging_activated:
            log.info(f"Logging path for socketCAN set to {self.log_path} ")
            self.logger = SocketCan2Trc(self.channel, str(self.log_path))
            self.logger.start()

    def _cc_close(self) -> None:
        """Close the current can bus channel and close the log handler."""
        self.bus.shutdown()
        self.bus = None

        if self.logging_activated:
            del self.logger
            self.logger = None

    def _cc_send(
        self, msg: MessageType, remote_id: int = None, raw: bool = False
    ) -> None:
        """Send a CAN message at the configured id.

        If remote_id parameter is not given take configured ones, in addition if
        raw is set to True take the msg parameter as it is otherwise parse it using
        test entity protocol format.

        :param msg: data to send
        :param remote_id: destination can id used
        :param raw: boolean use to select test entity protocol format

        """
        _data = msg

        if remote_id is None:
            remote_id = self.remote_id

        if not raw:
            _data = msg.serialize()

        can_msg = can.Message(
            arbitration_id=remote_id,
            data=_data,
            is_extended_id=self.is_extended_id,
            is_fd=self.is_fd,
            bitrate_switch=self.enable_brs,
        )
        self.bus.send(can_msg)

        log.debug(f"{self} sent CAN Message: {can_msg}, data: {_data}")

    def _cc_receive(
        self, timeout: float = 0.0001, raw: bool = False
    ) -> Union[Message, bytes, None]:
        """Receive a can message using configured filters.

        If raw parameter is set to True return received message as it is (bytes)
        otherwise test entity protocol format is used and Message class type is returned.

        :param timeout: timeout applied on reception
        :param raw: boolean use to select test entity protocol format

        :return: tuple containing the received data and the source can id
        """
        try:  # Catch bus errors & rcv.data errors when no messages where received
            received_msg = self.bus.recv(timeout=timeout or self.timeout)
            if received_msg is not None:
                frame_id = received_msg.arbitration_id
                payload = received_msg.data
                timestamp = received_msg.timestamp
                if not raw:
                    payload = Message.parse_packet(payload)
                log.debug(
                    "received CAN Message: {}, {}, {}".format(
                        frame_id, payload, timestamp
                    )
                )
                return payload, frame_id
            else:
                return None, None
        except can.CanError as can_error:
            log.debug(f"encountered can error: {can_error}")
            return None, None
        except Exception:
            log.exception(f"encountered error while receiving message via {self}")
            return None, None



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
CommunicationAuxiliary
**********************

:module: communication_auxiliary

:synopsis: Auxiliary used to send raw bytes via a connector instead of
    pykiso.Messages

.. currentmodule:: communication_auxiliary


"""

import logging

from pykiso import AuxiliaryInterface, CChannel, Message

log = logging.getLogger(__name__)


class CommunicationAuxiliary(AuxiliaryInterface):
    """Auxiliary used to send raw bytes via a connector instead of pykiso.Messages."""

    def __init__(self, com: CChannel, **kwargs):
        """Constructor.

        :param com: CChannel that supports raw communication
        """
        super().__init__(is_proxy_capable=True, **kwargs)
        self.channel = com

    def send_message(self, raw_msg: bytes) -> bool:
        """Send a raw message (bytes) via the communication channel.

        :param raw_msg: message to send

        :return: True if command was executed otherwise False
        """
        return self.run_command("send", raw_msg, timeout_in_s=None)

    def receive_message(
        self, blocking: bool = True, timeout_in_s: float = None
    ) -> bytes:
        """Receive a raw message.

        :param blocking: wait for message till timeout elapses?
        :param timeout_in_s: maximum time in second to wait for a response

        :returns: raw message
        """
        log.debug(
            f"retrieving message in {self} (blocking={blocking}, timeout={timeout_in_s})"
        )
        msg = self.wait_and_get_report(blocking=blocking, timeout_in_s=timeout_in_s)
        log.debug(f"retrieved message '{msg}' in {self}")
        return msg

    def _create_auxiliary_instance(self) -> bool:
        """Open the connector communication.

        :return: always True
        """
        state = False
        log.info("Create auxiliary instance")
        log.info("Enable channel")
        try:
            self.channel.open()
            state = True
        except Exception:
            log.exception("Unable to open channel communication")
            self.stop()
        return state

    def _delete_auxiliary_instance(self) -> bool:
        """Close the connector communication.

        :return: always True
        """
        log.info("Delete auxiliary instance")
        try:
            self.channel.close()
        except Exception:
            log.exception("Unable to close channel communication")
        return True

    def _run_command(self, cmd_message: bytes, cmd_data: bytes = None) -> bool:
        """Run the corresponding command.

        :param cmd_message: command type
        :param cmd_data: payload data to send over CChannel

        :return: True if command is executed otherwise False
        """
        if cmd_message == "send":
            try:
                self.channel.cc_send(msg=cmd_data, raw=True)
                return True
            except Exception:
                log.exception(
                    f"encountered error while sending message '{cmd_data}' to {self.channel}"
                )
        elif isinstance(cmd_message, Message):
            log.debug(f"ignored command '{cmd_message} in {self}'")
            return True
        else:
            log.warning(f"received unknown command '{cmd_message} in {self}'")
        return False

    def _abort_command(self) -> None:
        """No-op since we don't wait for ACKs"""
        pass

    def _receive_message(self, timeout_in_s: float) -> bytes:
        """No-op since it's handled in _run_command

        :param timeout_in_s: not used

        :return: received message
        """
        try:
            rcv_data = self.channel.cc_receive(timeout=0, raw=True)
            log.debug(f"received message '{rcv_data}' from {self.channel}")
            return rcv_data
        except Exception:
            log.exception(
                f"encountered error while receiving message via {self.channel}"
            )
            return None



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Example Auxiliary
*****************

:module: example_test_auxiliary

:synopsis: Example auxiliary that simulates a normal test run without
    ever doing anything.

.. currentmodule:: example_test_auxiliary


.. warning:: Still under test
"""
import logging

# Import the framework modules
from pykiso import AuxiliaryInterface, message

log = logging.getLogger(__name__)


class ExampleAuxiliary(AuxiliaryInterface):
    """Example of an auxiliary implementation."""

    def __init__(
        self,
        name=None,
        com=None,
        flash=None,
        **kwargs,
    ) -> None:
        """Constructor.

        :param name: Alias of the auxiliary instance
        :param com: Communication connector
        :param flash: flash connector
        """
        super().__init__(name=name, **kwargs)
        self.channel = com
        self.flash = flash
        self.options = kwargs
        log.debug(f"com is {com}")
        log.debug(f"flash is {flash}")

    def _create_auxiliary_instance(self) -> bool:
        """Open current associated channel and send a ping command.

        :return: if channel creation is successful and ping command is
            answered return True otherwise False
        """
        log.info("Create auxiliary instance")
        # Create an instance (e.g. flash it) -> Placeholder
        # Enable the communication with it
        log.info("Enable channel")
        self.channel.open()
        # Pingpong
        return_code = self._send_and_wait_ack(message.Message(), 0.5, 2)
        # Return output
        return return_code

    def _delete_auxiliary_instance(self) -> bool:
        """Close the connector communication.

        :return: always True
        """
        log.info("Delete auxiliary instance")
        # Close the communication with it
        self.channel.close()
        return True

    def _run_command(
        self, cmd_message: message.Message, cmd_data: bytes = None
    ) -> bool:
        """Run the corresponding command.

        :param cmd_message: command type
        :param cmd_data: payload data to send over CChannel

        :return: True if ACK is received otherwise False
        """
        log.debug("Send test request: {}".format(cmd_message))
        return self._send_and_wait_ack(cmd_message, 0.5, 2)

    def _abort_command(self) -> bool:
        """Send an abort command and reset the auxiliary if needed.

        :return: True if ACK is received otherwise False
        """
        log.info("Send abort request")
        # Try a soft abort
        msg = message.Message(
            message.MessageType.COMMAND, message.MessageCommandType.ABORT
        )  # TODO verify if this generation can happened somewhere else
        result = self._send_and_wait_ack(msg, 0.5, 2)
        # If not successful, do hard reset
        if result is False:
            self._delete_auxiliary_instance()
            self._create_auxiliary_instance()
        return result

    def _receive_message(self, timeout_in_s: float) -> message.Message:
        """Get message from the device under test.

        :param timeout_in_s: Time in ms to wait for one try

        :return: receive message
        """
        # Read message on the channel
        received_message = self.channel.cc_receive(timeout_in_s)
        if received_message is not None:
            # Send ack
            self.channel._cc_send(
                msg=received_message.generate_ack_message(message.MessageAckType.ACK)
            )
        # Return message
        return received_message

    def _send_and_wait_ack(
        self, message_to_send: message.Message, timeout: float, tries: int
    ) -> bool:
        """Send via the channel a message and wait for an acknowledgement.

        :param message_to_send: Message you want to send out
        :param timeout: Time in ms to wait for one try
        :param tries: Number of tries to send the message

        :return: True - Ack, False - Nack
        """
        number_of_tries = 0
        result = False
        while number_of_tries < tries:
            log.debug("Run send try n:" + str(number_of_tries))
            # Increase number of tries
            number_of_tries += 1
            # Send the message
            self.channel.cc_send(msg=message_to_send)
            # Wait until we get something back
            received_message = self.channel.cc_receive(timeout)
            # Check the outcome
            if received_message is None:
                continue  # Next try will occur
            else:
                if message_to_send.check_if_ack_message_is_matching(received_message):
                    log.debug("{} sent".format(message_to_send))
                    result = True
                    break
                else:
                    log.warning(
                        "Received {} not matching {}!".format(
                            received_message, message_to_send
                        )
                    )
                    continue  # A NACK got received # TODO later on we should have "not supported" and ignore it than.
        # Must be ">" because number_of_tries is increased at the beginning of the while loop above
        if number_of_tries > tries:
            log.warning(
                f"Exceeded the number of tries ({tries}) for sending the message"
            )
        return result



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Can Communication Channel using PCAN hardware
*********************************************

:module: cc_pcan_can

:synopsis: CChannel implementation for CAN(fd) using PCAN API from python-can

.. currentmodule:: cc_pcan_can

"""

import logging
from pathlib import Path
from typing import Union

import can
import can.bus
import can.interfaces.pcan.basic as PCANBasic

from pykiso import CChannel, Message

MessageType = Union[Message, bytes]

log = logging.getLogger(__name__)


class CCPCanCan(CChannel):
    """CAN FD channel-adapter."""

    def __init__(
        self,
        interface: str = "pcan",
        channel: str = "PCAN_USBBUS1",
        state: str = "ACTIVE",
        bitrate: int = 500000,
        is_fd: bool = True,
        enable_brs: bool = False,
        f_clock_mhz: int = 80,
        nom_brp: int = 2,
        nom_tseg1: int = 63,
        nom_tseg2: int = 16,
        nom_sjw: int = 16,
        data_brp: int = 4,
        data_tseg1: int = 7,
        data_tseg2: int = 2,
        data_sjw: int = 2,
        is_extended_id: bool = False,
        remote_id: int = None,
        can_filters: list = None,
        logging_activated: bool = True,
        **kwargs,
    ):
        """Initialize can channel settings.

        :param interface: python-can interface modules used
        :param channel: the can interface name
        :param state: BusState of the channel
        :param bitrate: Bitrate of channel in bit/s,ignored if using CanFD
        :param is_fd: Should the Bus be initialized in CAN-FD mode
        :param enable_brs: sets the bitrate_switch flag to use higher transmission speed
        :param f_clock_mhz:  Clock rate in MHz
        :param nom_brp: Clock prescaler for nominal time quantum
        :param nom_tseg1: Time segment 1 for nominal bit rate, that is,
            the number of quanta from the Sync Segment to the sampling point
        :param nom_tseg2: Time segment 2 for nominal bit rate,
            that is, the number of quanta from the sampling point to the end of the bit
        :param nom_sjw: Synchronization Jump Width for nominal bit rate.
            Decides the maximum number of time quanta that the controller
            can resynchronize every bit
        :param data_brp: Clock prescaler for fast data time quantum
        :param data_tseg1: Time segment 1 for fast data bit rate, that is,
            the number of quanta from the Sync Segment to the sampling point
        :param data_tseg2: Time segment 2 for fast data bit rate, that is,
            the number of quanta from the sampling point to the end of the bit.
            In the range (1..16)
        :param data_sjw: Synchronization Jump Width for fast data bit rate
        :param is_extended_id: This flag controls the size of the arbitration_id field
        :param remote_id: id used for transmission
        :param can_filters: iterable used to filter can id on reception
        :param logging_activated: boolean used to disable logfile creation
        """
        super().__init__(**kwargs)
        self.interface = interface
        self.channel = channel
        self.state = can.bus.BusState[state.upper()]
        self.bitrate = bitrate
        self.is_fd = is_fd
        self.enable_brs = enable_brs
        self.f_clock_mhz = f_clock_mhz
        self.nom_brp = nom_brp
        self.nom_tseg1 = nom_tseg1
        self.nom_tseg2 = nom_tseg2
        self.nom_sjw = nom_sjw
        self.data_brp = data_brp
        self.data_tseg1 = data_tseg1
        self.data_tseg2 = data_tseg2
        self.data_sjw = data_sjw
        self.is_extended_id = is_extended_id
        self.remote_id = remote_id
        self.can_filters = can_filters
        self.bus = None
        self.logging_activated = logging_activated
        self.raw_pcan_interface = None
        self.logging_path = None
        # Set a timeout to send the signal to the GIL to change thread.
        # In case of a multi-threading system, all tasks will be called one after the other.
        self.timeout = 1e-6
        """
        Extract the base logging directory from the logging module, so we can
        create our logging folder in the correct place.
        logging_path will be set to the parent directory of the logfile which
        is set in the logging module + /raw/PCAN
        If an AttributeError occurs, the logging module is not set to log into
        a file.
        In this case we set the logging_path to None and will just log into a
        generic logfile in the current working directory, which will be
        overwritten every time, a log is initiated.
        """
        if self.logging_activated:
            try:
                self.logging_path = (
                    Path(log.getLoggerClass().root.handlers[0].baseFilename)
                    .absolute()
                    .parent
                    / "raw/PCAN/"
                )
            except AttributeError:
                log.warning("Logging path of the logging module not set to file")

        log.info(f"Logging path for PCAN set to {self.logging_path}")

        if self.enable_brs and not self.is_fd:
            log.warning(
                "Bitrate switch will have no effect because option is_fd is set to false."
            )

    def _cc_open(self) -> None:
        """Open a can bus channel, set filters for reception and activate PCAN log."""
        self.bus = can.interface.Bus(
            interface=self.interface,
            channel=self.channel,
            state=self.state,
            bitrate=self.bitrate,
            fd=self.is_fd,
            f_clock_mhz=self.f_clock_mhz,
            nom_brp=self.nom_brp,
            nom_tseg1=self.nom_tseg1,
            nom_tseg2=self.nom_tseg2,
            nom_sjw=self.nom_sjw,
            data_brp=self.data_brp,
            data_tseg1=self.data_tseg1,
            data_tseg2=self.data_tseg2,
            data_sjw=self.data_sjw,
            can_filters=self.can_filters,
        )

        if self.logging_activated and self.raw_pcan_interface is None:
            self.raw_pcan_interface = PCANBasic.PCANBasic()
            self._pcan_configure_trace()

    def _pcan_configure_trace(self) -> None:
        """Configure PCAN dongle to create a trace file.

        If self.logging_path is set, this path will be created, if it does not
        exist and the logfile will be placed there.
        Otherwise it will be logged to the current working directory if a
        default filename, which will be overwritten in successive calls.
        If an error occurs, the trace will not be started and the error logged.
        No exception is thrown in this case.
        """
        pcan_channel = getattr(PCANBasic, self.channel)
        if self.logging_path is None:
            pcan_path_argument = PCANBasic.TRACE_FILE_OVERWRITE
        else:
            pcan_path_argument = PCANBasic.TRACE_FILE_DATE | PCANBasic.TRACE_FILE_TIME
        try:
            if self.logging_path is not None:
                if not Path(self.logging_path).exists():
                    Path(self.logging_path).mkdir(parents=True, exist_ok=True)
                    log.info(f"Path {self.logging_path} created")
                self._pcan_set_value(
                    pcan_channel,
                    PCANBasic.PCAN_TRACE_LOCATION,
                    bytes(self.logging_path),
                )
                log.info(
                    f"Tracefile path in PCAN device configured to {self.logging_path}"
                )
            self._pcan_set_value(
                pcan_channel,
                PCANBasic.PCAN_TRACE_CONFIGURE,
                pcan_path_argument,
            )
            log.info("Tracefile configured")
            self._pcan_set_value(
                pcan_channel, PCANBasic.PCAN_TRACE_STATUS, PCANBasic.PCAN_PARAMETER_ON
            )
            log.info("Trace activated")
        except RuntimeError:
            log.error(f"Logging for {self.channel} not activated")
        except OSError as e:
            log.error(f"Can not create log folder for PCAN logs: {e}")
            log.error(f"Logging for {self.channel} not activated")

    def _pcan_set_value(self, channel, parameter, buffer) -> None:
        """Set a value in the PCAN api.

        If this is not successful, a RuntimeError is returned, as well as the
        PCAN error text is logged, if possible.

        :param channel: Channel for PCANBasic.SetValue
        :param parameter: Parameter for PCANBasic.SetValue
        :param buffer: Buffer for PCANBasic.SetValue

        :raises RuntimeError: Raised if the function is not successful
        """
        try:
            result = self.raw_pcan_interface.SetValue(
                channel,
                parameter,
                buffer,
            )
        except Exception as e:
            log.error(f"Exception in call to SetValue: {e}")
            raise RuntimeError("Error configuring logging on PCAN")
        else:
            if result != PCANBasic.PCAN_ERROR_OK:
                _, error_msg = self.raw_pcan_interface.GetErrorText(result)
                log.error(error_msg)
                raise RuntimeError(f"Error configuring logging on PCAN: {result}")

    def _cc_close(self) -> None:
        """Close the current can bus channel and uninitialize PCAN handle."""
        self.bus.shutdown()
        self.bus = None
        if self.logging_activated:
            try:
                result = self.raw_pcan_interface.Uninitialize(PCANBasic.PCAN_NONEBUS)
            except Exception as e:
                log.exception(f"Error in call to Uninitialize: {e}")
            else:
                if result != PCANBasic.PCAN_ERROR_OK:
                    _, error_msg = self.raw_pcan_interface.GetErrorText(result)
                    log.error(error_msg)
            finally:
                self.raw_pcan_interface = None

    def _cc_send(
        self, msg: MessageType, remote_id: int = None, raw: bool = False
    ) -> None:
        """Send a CAN message at the configured id.

        If remote_id parameter is not given take configured ones, in addition if
        raw is set to True take the msg parameter as it is otherwise parse it using
        test entity protocol format.

        :param msg: data to send
        :param remote_id: destination can id used
        :param raw: boolean use to select test entity protocol format

        """
        _data = msg

        if remote_id is None:
            remote_id = self.remote_id

        if not raw:
            _data = msg.serialize()

        can_msg = can.Message(
            arbitration_id=remote_id,
            data=_data,
            is_extended_id=self.is_extended_id,
            is_fd=self.is_fd,
            bitrate_switch=self.enable_brs,
        )
        self.bus.send(can_msg)

        log.debug(f"{self} sent CAN Message: {can_msg}, data: {_data}")

    def _cc_receive(
        self, timeout: float = 0.0001, raw: bool = False
    ) -> Union[Message, bytes, None]:
        """Receive a can message using configured filters.

        If raw parameter is set to True return received message as it is (bytes)
        otherwise test entity protocol format is used and Message class type is returned.

        :param timeout: timeout applied on reception
        :param raw: boolean use to select test entity protocol format

        :return: tuple containing the received data and the source can id
        """
        try:  # Catch bus errors & rcv.data errors when no messages where received
            received_msg = self.bus.recv(timeout=timeout or self.timeout)
            if received_msg is not None:
                frame_id = received_msg.arbitration_id
                payload = received_msg.data
                timestamp = received_msg.timestamp
                if not raw:
                    payload = Message.parse_packet(payload)
                log.debug(f"received CAN Message: {frame_id}, {payload}, {timestamp}")
                return payload, frame_id
            else:
                return None, None
        except can.CanError as can_error:
            log.debug(f"encountered can error: {can_error}")
            return None, None
        except Exception:
            log.exception(f"encountered error while receiving message via {self}")
            return None, None



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Test Execution
**************

:module: test_execution

:synopsis: Execute a test environment based on the supplied configuration.

.. currentmodule:: test_execute

.. note::
    1. Glob a list of test-suite folders
    2. Generate a list of test-suites with a list of test-cases
    3. Loop per suite
    4. Gather result
"""
import enum
import itertools
import logging
import time
import unittest
from pathlib import Path
from typing import Dict, Optional

import xmlrunner

from ..exceptions import AuxiliaryCreationError
from . import test_suite
from .test_result import BannerTestResult
from .test_xml_result import XmlTestResult

log = logging.getLogger(__name__)


@enum.unique
class ExitCode(enum.IntEnum):
    """List of possible exit codes"""

    ALL_TESTS_SUCCEEDED = 0
    ONE_OR_MORE_TESTS_FAILED = 1
    ONE_OR_MORE_TESTS_RAISED_UNEXPECTED_EXCEPTION = 2
    ONE_OR_MORE_TESTS_FAILED_AND_RAISED_UNEXPECTED_EXCEPTION = 3
    AUXILIARY_CREATION_FAILED = 4


def create_test_suite(test_suite_dict: Dict) -> test_suite.BasicTestSuite:
    """create a test suite based on the config dict

    :param test_suite_dict: dict created from config with keys 'suite_dir',
        'test_filter_pattern', 'test_suite_id'
    """
    return test_suite.BasicTestSuite(
        modules_to_add_dir=test_suite_dict["suite_dir"],
        test_filter_pattern=test_suite_dict["test_filter_pattern"],
        test_suite_id=test_suite_dict["test_suite_id"],
        args=[],
        kwargs={},
    )


def apply_variant_filter(
    all_tests_to_run: dict, variants: tuple, branch_levels: tuple
) -> None:
    """Filter the test cases based on the variant string.

    :param all_tests_to_run: a dict containing all testsuites and testcases
    :param variants: encapsulate user's variant choices
    :param branch_levels: encapsulate user's branch level choices
    """

    def tag_present(tag_list: tuple, tc_tags: list) -> bool:
        """Determine if at least branch_level or variant tag is present
        at test fixture decorator level.

        :param tag_list: encapsulate user's variant choices
        :param tc_tags: encapsulate all available tags on test case
            decorator level

        :return: True if a variant or branch_level is found
            otherwise False
        """
        for tag in tag_list:
            if tag in tc_tags:
                return True
        else:
            return False

    def both_present(variants: tuple, branches: tuple, tc_tags: list) -> bool:
        """Determine if the couple branch_level and variant is present
        at test fixture decorator level.

        :param variants: encapsulate user's variant choices
        :param branches: encapsulate user's branch level choices
        :param tc_tags: encapsulate all available tags on test case
            decorator level

        :return: True if a couple variant/branch_level is found
            otherwise False
        """
        for variant in variants:
            for branch in branches:
                if variant in tc_tags and branch in tc_tags:
                    return True
        else:
            return False

    base_suite = test_suite.flatten(all_tests_to_run)
    both_given = variants and branch_levels

    for tc in base_suite:
        # if variant at decorator level is not given just run it
        if tc.variant is not None:
            # extract variant and branch from test fixture
            tc_variants = tc.variant.get("variant", list())
            tc_branches = tc.variant.get("branch_level", list())
            tc_tags = list(itertools.chain(tc_variants, tc_branches))

            # if user gives both branch and variant filter using couple
            # (variant, branch_level)
            if both_given and both_present(variants, branch_levels, tc_tags):
                continue
            # user gives only variant param, filter using only variant
            elif not both_given and tag_present(variants, tc_tags):
                continue
            # user gives only branch param, filter using only branch
            elif not both_given and tag_present(branch_levels, tc_tags):
                continue
            # the test is not intended to be run skip it
            else:
                tc.setUp = lambda: "setup_skipped"
                setattr(
                    tc,
                    tc._testMethodName,
                    lambda: tc.skipTest("skipped due to non-matching variant value"),
                )
                tc.tearDown = lambda: "tearDown_skipped"


def failure_and_error_handling(result: unittest.TestResult) -> int:
    """provide necessary information to Jenkins if an error occur during tests execution

    :param result: encapsulate all test results from the current run

    :return: an ExitCode object
    """
    failed, errored = result.failures, result.errors
    if failed and errored:
        exit_code = ExitCode.ONE_OR_MORE_TESTS_FAILED_AND_RAISED_UNEXPECTED_EXCEPTION
    elif failed:
        exit_code = ExitCode.ONE_OR_MORE_TESTS_FAILED
    elif errored:
        exit_code = ExitCode.ONE_OR_MORE_TESTS_RAISED_UNEXPECTED_EXCEPTION
    else:
        exit_code = ExitCode.ALL_TESTS_SUCCEEDED
    return exit_code


def execute(
    config: Dict,
    report_type: str = "text",
    variants: Optional[tuple] = None,
    branch_levels: Optional[tuple] = None,
    pattern_inject: Optional[str] = None,
    failfast: bool = False,
) -> int:
    """create test environment base on config

    :param config: dict from converted YAML config file
    :param report_type: str to set the type of report wanted, i.e. test
        or junit
    :param variants: encapsulate user's variant choices
    :param branch_levels: encapsulate user's branch level choices
    :param pattern_inject: optional pattern that will override
        test_filter_pattern for all suites. Used in test development to
        run specific tests.
    :param failfast: Stop the test run on the first error or failure

    :return: exit code corresponding to the actual tets exeuciton run
        state(tests failed, unexpected exception...)
    """
    exit_code = ExitCode.ONE_OR_MORE_TESTS_RAISED_UNEXPECTED_EXCEPTION
    is_raised = False

    try:
        list_of_test_suites = []
        for test_suite_configuration in config["test_suite_list"]:
            try:
                if pattern_inject is not None:
                    test_suite_configuration["test_filter_pattern"] = pattern_inject
                list_of_test_suites.append(create_test_suite(test_suite_configuration))
            except BaseException:
                is_raised = True
                break

        # Collect all the tests in one global test suite
        all_tests_to_run = unittest.TestSuite(list_of_test_suites)
        if variants or branch_levels:
            apply_variant_filter(all_tests_to_run, variants, branch_levels)
        # TestRunner selection: generate or not a junit report. Start the tests and publish the results
        if report_type == "junit":
            junit_report_name = time.strftime("TEST-pykiso-%Y-%m-%d_%H-%M-%S.xml")
            project_folder = Path.cwd()
            reports_path = project_folder / "reports"
            junit_report_path = reports_path / junit_report_name
            reports_path.mkdir(exist_ok=True)
            with open(junit_report_path, "wb") as junit_output:
                test_runner = xmlrunner.XMLTestRunner(
                    output=junit_output,
                    resultclass=XmlTestResult,
                    failfast=failfast,
                )
                result = test_runner.run(all_tests_to_run)
        else:
            test_runner = unittest.TextTestRunner(
                resultclass=BannerTestResult, failfast=failfast
            )
            result = test_runner.run(all_tests_to_run)

        # if an exception is raised during test suite collections at least
        # return exit code ONE_OR_MORE_TESTS_RAISED_UNEXPECTED_EXCEPTION
        if is_raised:
            exit_code = ExitCode.ONE_OR_MORE_TESTS_RAISED_UNEXPECTED_EXCEPTION
        else:
            exit_code = failure_and_error_handling(result)
    except AuxiliaryCreationError:
        exit_code = ExitCode.AUXILIARY_CREATION_FAILED
    except KeyboardInterrupt:
        log.exception("Keyboard Interrupt detected")
        exit_code = ExitCode.ONE_OR_MORE_TESTS_RAISED_UNEXPECTED_EXCEPTION
    except Exception:
        log.exception(f'Issue detected in the test-suite: {config["test_suite_list"]}!')
        exit_code = ExitCode.ONE_OR_MORE_TESTS_RAISED_UNEXPECTED_EXCEPTION
    finally:
        return int(exit_code)



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Testapp binding
***************

:module: dut_auxiliary

:synopsis: implementation of existing DUTAuxiliary for Robot
    framework usage.

.. currentmodule:: dut_auxiliary

"""

from typing import List

from robot.api import logger
from robot.api.deco import keyword, library

from pykiso.interfaces.thread_auxiliary import AuxiliaryInterface
from pykiso.message import MessageCommandType, MessageReportType, MessageType
from pykiso.test_coordinator.test_message_handler import (
    handle_basic_interaction,
)

from ..auxiliaries.dut_auxiliary import DUTAuxiliary as DutAux
from .aux_interface import RobotAuxInterface


class TestEntity:
    """Dummy Class to use handle_basic_interaction from test_message_handler."""

    def __init__(
        self,
        test_suite_id: int,
        test_case_id: int,
        aux_list: List[AuxiliaryInterface],
    ):
        """Initialize generic test-case

        :param test_suite_id: test suite identification number
        :param test_case_id: test case identification number
        :param aux_list: list of used aux_list"""

        self.test_suite_id = test_suite_id
        self.test_case_id = test_case_id
        self.test_auxiliary_list = aux_list

    def cleanup_and_skip(self, aux: AuxiliaryInterface, info_to_print: str):
        """Cleanup auxiliary and log reasons.

        :param aux: corresponding auxiliary to abort
        :param info_to_print: A message you want to print while cleaning up the test
        """
        logger.error(info_to_print)

        # Send aborts to corresponding auxiliary
        if aux.abort_command() is not True:
            logger.error(f"Error occurred during abort command on auxiliary {aux}")


@library(version="0.0.1")
class DUTAuxiliary(RobotAuxInterface):
    """Robot library to control the TestApp on the DUT"""

    ROBOT_LIBRARY_SCOPE = "SUITE"

    def __init__(self):
        """Initialize attributes."""
        super().__init__(aux_type=DutAux)

    @keyword(name="Test App")
    def test_app_run(
        self,
        command_type: str,
        test_suite_id: int,
        test_case_id: int,
        aux_list: List[str],
        timeout_cmd: int = 5,
        timeout_resp: int = 5,
    ) -> None:
        """Handle default communication mechanism between test manager and device under test.

        :param command_type: message command sub-type (TEST_SECTION_SETUP , TEST_SECTION_RUN, ...)
        :param test_suite_id: select test suite id on dut
        :param test_case_id: select test case id on dut
        :param aux_list: List of selected auxiliary
        :param timeout_cmd: timeout in seconds for auxiliary run_command
        :param timeout_resp: timeout in seconds for auxiliary wait_and_get_report
        """

        auxiliaries = [self._get_aux(aux) for aux in aux_list]

        test_entity = TestEntity(test_suite_id, test_case_id, auxiliaries)

        with handle_basic_interaction(
            test_entity,
            MessageCommandType[command_type],
            timeout_cmd,
            timeout_resp,
        ) as report_infos:

            for _, report_msg, log_level_func, log_msg in report_infos:
                log_level_func(log_msg)

                if report_msg.get_message_type() == MessageType.REPORT:

                    if not report_msg.sub_type == MessageReportType.TEST_PASS:
                        # Raise assertion to make test red in robotframework
                        raise AssertionError(f"Test Failed:{log_msg}")



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Config Registry
***************

:module: config_registry

:synopsis: register auxiliaries and connectors to provide them for import.

.. currentmodule:: config_registry

"""

from typing import Any

from .dynamic_loader import DynamicImportLinker


class ConfigRegistry:
    """Register auxiliaries with connectors to provide systemwide import statements."""

    _linker = None

    @classmethod
    def register_aux_con(cls, config: dict) -> None:
        """Create import hooks. Register auxiliaries and connectors.

        :param config: dictionary containing yaml configuration content
        """
        ConfigRegistry._linker = DynamicImportLinker()
        ConfigRegistry._linker.install()
        for connector, con_details in config["connectors"].items():
            cfg = con_details.get("config") or dict()
            ConfigRegistry._linker.provide_connector(
                connector, con_details["type"], **cfg
            )
        for auxiliary, aux_details in config["auxiliaries"].items():
            cfg = aux_details.get("config") or dict()
            ConfigRegistry._linker.provide_auxiliary(
                auxiliary,
                aux_details["type"],
                aux_cons=aux_details.get("connectors") or dict(),
                **cfg,
            )

    @classmethod
    def delete_aux_con(cls) -> None:
        """deregister the import hooks, close all running threads, delete all instances."""
        ConfigRegistry._linker.uninstall()

    @classmethod
    def get_all_auxes(cls) -> dict:
        """Return all auxiliaires instances and alias

        :return: dictionary with alias as keys and instances as values
        """
        return ConfigRegistry._linker._aux_cache.instances

    @classmethod
    def get_auxes_by_type(cls, aux_type: Any) -> dict:
        """Return all auxiliaries who match a specific type.

        :param aux_type: auxiliary class type (DUTAuxiliary,
            CommunicationAuxiliary...)

        :return: dictionary with alias as keys and instances as values
        """
        all_auxes = ConfigRegistry._linker._aux_cache.instances
        return {
            alias: inst
            for alias, inst in all_auxes.items()
            if isinstance(inst, aux_type)
        }

    @classmethod
    def get_aux_config(cls, name: str) -> dict:
        """Return the registered auxiliary configuration based on his
        name.

        :param name: auxiliary alias

        :return: auxiliary's configuration (yaml content)
        """
        return ConfigRegistry._linker._aux_cache.configs[name]

    @classmethod
    def get_auxes_alias(cls) -> list:
        """return all created auxiliaries alias.

        :return: list containing all auxiliaries alias
        """
        return [alias for alias in ConfigRegistry._linker._aux_cache.connectors.keys()]



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Text Test Result with banners
*****************************

:module: test_result

:synopsis: implement BannerTestResult to show detailed test execution
wrapped in banners.

.. currentmodule:: test_result
"""

import textwrap
from shutil import get_terminal_size
from typing import List, Optional, TextIO, Union
from unittest import TestCase, TestResult, TextTestResult


class BannerTestResult(TextTestResult):
    """TextTestResult subclass showing results wrapped in banners."""

    def __init__(self, stream: TextIO, descriptions: bool, verbosity: int):
        """Constructor. Initialize TextTestResult and the banner's width.

        The banner's width is set to the terminal size. In the case
        where this fails the fallback width corresponds to the default
        width of a Jenkins "console".

        :param stream: stream to print the result information
            (default: stderr)
        :param descriptions: unused (required for TextTestResult)
        :param verbosity: unused (required for TextTestResult)
        """
        super().__init__(stream, descriptions, verbosity)
        # disable TextTestResult printing
        self.dots = False
        self.verbosity = False
        self._error_occurred = False
        # fallback is the default width in Jenkins
        size = get_terminal_size(fallback=(150, 24))
        # avoid border effects due to newlines
        self.width = size.columns - 1

    def _banner(
        self, text: Union[List, str], width: Optional[int] = None, sym: str = "#"
    ) -> str:
        """Format text as a banner.

        Works with multiline strings (either as a string containing
        newlines or split into a list with one entry per line).

        :param text: text to format
        :paran width: width of the banner
        :param sym: symbol used to compose the banner

        :return: the text enclosed in a banner
        """
        width = width or self.width
        bar = sym * width
        if isinstance(text, str):
            text = text.split("\n")
        if isinstance(text, list):
            text = "\n".join(f"{sym} {line: <{width-4}} {sym}" for line in text)
        banner = f"{bar}\n{text}\n{bar}\n"
        return banner

    def getDescription(self, test: TestCase) -> str:
        """Return the entire test method docstring.

        :param test: running testcase

        :return: the wrapped docstring
        """
        doc = ""
        if test._testMethodDoc:
            doc = "\n"
            for line in test._testMethodDoc.splitlines():
                doc += "\n" + textwrap.fill(line.strip(), width=100)
        return doc

    def startTest(self, test: TestCase) -> None:
        """Print a banner containing the test information and the test
        method docstring when starting a test case.

        :param test: running testcase
        """
        TestResult.startTest(self, test)
        self._error_occurred = False
        top_str = "RUNNING TEST: "
        module_name = test.__module__
        test_name = str(test)
        addendum = ""
        doc = self.getDescription(test)
        if len(module_name + test_name) < self.width - len(top_str):
            test_name = f"{module_name}.{test_name}"
        else:
            addendum += f"\nmodule: {module_name}"
        top_str += f"{test_name}{addendum}{doc}"
        top_banner = self._banner(top_str)
        self.stream.write(top_banner)
        self.stream.flush()

    def stopTest(self, test: TestCase) -> None:
        """Print a banner containing the test information and its result.

        :param test: running testcase
        """
        result = "Failed" if self._error_occurred else "Passed"
        bot_str = f"END OF TEST: {test}"
        result_str = f"  ->  {result}"
        if len(bot_str + result_str) < self.width:
            bot_str += result_str
        else:
            bot_str += "\n" + result_str
        bot_banner = self._banner(bot_str) + "\n"
        self.stream.write(bot_banner)
        self.stream.flush()
        TestResult.stopTest(self, test)

    def addFailure(self, test: TestCase, err: tuple) -> None:
        """Set the error flag when a failure occurs in order to get the
        individual test case result.

        :param test: running testcase
        :param err: tuple returned by sys.exc_info
        """
        TestResult.addFailure(self, test, err)
        self._error_occurred = True

    def addError(self, test: TestCase, err: tuple) -> None:
        """Set the error flag when an error occurs in order to get the
        individual test case result.

        :param test: running testcase
        :param err: tuple returned by sys.exc_info
        """
        TestResult.addError(self, test, err)
        self._error_occurred = True

    def printErrorList(self, flavour: str, errors: List[tuple]):
        """Print all errors at the end of the whole tests execution
        Overwrite the unit-test function

        :param flavour: failure reason
        :param errors: list of failed tests with their error message
        """
        for test, err in errors:
            self.stream.writeln(self.separator1)
            self.stream.writeln("%s" % test)
            self.stream.writeln("%s" % self.getDescription(test))
            self.stream.writeln("%s: %s" % (flavour, err))



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Generic Test
************

:module: test_case

:synopsis: Basic extensible implementation of a TestCase.

.. currentmodule:: test_case

.. note:: TODO later on will inherit from a metaclass to get the id parameters
"""

import functools
import logging
import unittest
from typing import List, Optional, Type, Union

from .. import message
from ..cli import get_logging_options, initialize_logging
from ..interfaces.thread_auxiliary import AuxiliaryInterface
from .test_message_handler import test_app_interaction

log = logging.getLogger(__name__)


def retry_test_case(
    max_try: int = 2,
    rerun_setup: bool = False,
    rerun_teardown: bool = False,
    stability_test: bool = False,
):
    """Decorator: retry mechanism for testCase.

    The aim is to cover the 2 following cases:

        - Unstable test : get the test pass within the {max_try} attempt

        - Stability test : run {max_try} time the test expecting no error

    The **retry_test_case** comes with the possibility to re-run the setUp and
    tearDown methods automatically.

    :param max_try: maximum number of try to get the test pass.
    :param rerun_setup: call the "setUp" method of the test.
    :param rerun_teardown: call the "tearDown" method of the test.
    :param stability_test: run {max_try} time the test and raise an exception if an error occurs.

    :return: None, a testCase is not supposed to return anything.
    :raise Exception: if stability_test, the exception that occurred during the execution; if
        not stability_test, the exception that occurred at the last try.
    """

    def decorator(func):
        @functools.wraps(func)
        def func_wrapper(self) -> None:
            # track the current execution for logging
            current_execution = None

            for retry_nb in range(1, max_try + 1):
                try:
                    # by the 2nd attempt, end the test with the teardown and start with setUp
                    if retry_nb > 1:
                        if rerun_teardown:
                            current_execution = self.tearDown
                            self.tearDown()
                        if rerun_setup:
                            current_execution = self.setUp
                            self.setUp()

                    # run the method (eg: test_run(self))
                    current_execution = func
                    func(self)
                    if not stability_test:
                        break
                    else:
                        # Clearly separate tests
                        log.info(
                            f">>>>>>>>>> Stability test {retry_nb}/{max_try} succeed <<<<<<<<<<"
                        )

                except Exception as e:
                    # log: test_name (class), method (setUp, test_run, tearDown) and the error.
                    log.warning(
                        f"{self.__class__.__name__}.{current_execution.__name__} failed with exception: {e}."
                    )

                    # raise the exception that occurred during the latest attempt
                    if retry_nb == max_try or stability_test:
                        log.error(
                            f">>>>>>>>>> Test {retry_nb}/{max_try} failed <<<<<<<<<<"
                        )
                        raise e

                    # print counter only after failing test to avoid spamming the console
                    log.info(f">>>>>>>>>> Attempt: {retry_nb +1}/{max_try} <<<<<<<<<<")

        return func_wrapper

    return decorator


class BasicTest(unittest.TestCase):
    """Base for test-cases."""

    response_timeout: int = 10

    def __init__(
        self,
        test_suite_id: int,
        test_case_id: int,
        aux_list: Union[List[AuxiliaryInterface], None],
        setup_timeout: Union[int, None],
        run_timeout: Union[int, None],
        teardown_timeout: Union[int, None],
        test_ids: Union[dict, None],
        variant: Union[list, None],
        args: tuple,
        kwargs: dict,
    ):
        """Initialize generic test-case.

        :param test_suite_id: test suite identification number
        :param test_case_id: test case identification number
        :param aux_list: list of used auxiliaries
        :param setup_timeout: maximum time (in seconds) used to wait
            for a report during setup execution
        :param run_timeout: maximum time (in seconds) used to wait for
            a report during test_run execution
        :param teardown_timeout: the maximum time (in seconds) used to
            wait for a report during teardown execution
        :param test_ids: jama references to get the coverage
            eg: {"Component1": ["Req1", "Req2"], "Component2": ["Req3"]}
        :param variant: string that allows the user to execute a subset of tests
        """
        # Initialize base class
        super().__init__(*args, **kwargs)
        # Save list of test auxiliaries to use (already initialize)
        self.test_auxiliary_list = aux_list or []
        # Save test information
        self.test_suite_id = test_suite_id
        self.test_case_id = test_case_id
        self.setup_timeout = setup_timeout or BasicTest.response_timeout
        self.run_timeout = run_timeout or BasicTest.response_timeout
        self.teardown_timeout = teardown_timeout or BasicTest.response_timeout
        self.test_ids = test_ids
        self.variant = variant

    def cleanup_and_skip(self, aux: AuxiliaryInterface, info_to_print: str) -> None:
        """Cleanup auxiliary and log reasons.

        :param aux: corresponding auxiliary to abort
        :param info_to_print: A message you want to print while cleaning up the test
        """
        # Log error message
        log.critical(info_to_print)

        # Send aborts to corresponding auxiliary
        if aux.abort_command() is not True:
            log.critical(f"Error occurred during abort command on auxiliary {aux}")

        self.fail(info_to_print)

    @classmethod
    def setUpClass(cls) -> None:
        """A class method called before tests in an individual class are
        run.

        This implementation is only mandatory to enable logging in junit
        report. The logging configuration has to be call inside test
        runner run, otherwise stdout is never caught.
        """
        options = get_logging_options()
        if options.report_type == "junit":
            initialize_logging(None, options.log_level, options.report_type)

    @test_app_interaction(
        message_type=message.MessageCommandType.TEST_CASE_SETUP, timeout_cmd=5
    )
    def setUp(self) -> None:
        """Hook method for constructing the test fixture."""
        pass

    @test_app_interaction(
        message_type=message.MessageCommandType.TEST_CASE_RUN, timeout_cmd=5
    )
    def test_run(self) -> None:
        """Hook method from unittest in order to execute test case."""
        pass

    @test_app_interaction(
        message_type=message.MessageCommandType.TEST_CASE_TEARDOWN, timeout_cmd=5
    )
    def tearDown(self) -> None:
        """Hook method for deconstructing the test fixture after testing it."""
        pass


def define_test_parameters(
    suite_id: int = 0,
    case_id: int = 0,
    aux_list: List[AuxiliaryInterface] = None,
    setup_timeout: Optional[int] = None,
    run_timeout: Optional[int] = None,
    teardown_timeout: Optional[int] = None,
    test_ids: Optional[dict] = None,
    variant: Optional[dict] = None,
):
    """Decorator to fill out test parameters of the BasicTest automatically."""

    def generate_modified_class(DecoratedClass):
        """Generates the same class but with the test IDs already filled.
        It works as a partially filled-out call to the __init__ method.
        """

        class NewClass(DecoratedClass):
            """Modified {DecoratedClass.__name__}, with the __init__ method
            already filled out with the following test-parameters:

            Suite ID:    {suite_id}
            Case ID:     {case_id}
            Auxiliaries: {auxes}
            setup_timeout: {setup_timeout}
            run_timeout: {run_timeout}
            teardown_timeout: {teardown_timeout}
            test_ids: {test_ids}
            variant: {variant}
            """

            @functools.wraps(DecoratedClass.__init__)
            def __init__(self, *args, **kwargs):
                super(NewClass, self).__init__(
                    suite_id,
                    case_id,
                    aux_list,
                    setup_timeout,
                    run_timeout,
                    teardown_timeout,
                    test_ids,
                    variant,
                    args,
                    kwargs,
                )

        NewClass.__doc__ = NewClass.__doc__.format(
            DecoratedClass=DecoratedClass,
            suite_id=suite_id,
            case_id=case_id,
            auxes=[aux.__class__.__name__ for aux in aux_list or []],
            setup_timeout=setup_timeout,
            run_timeout=run_timeout,
            teardown_timeout=teardown_timeout,
            test_ids=test_ids,
            variant=variant,
        )
        # Used to display the current test module in the test result
        NewClass.__module__ = DecoratedClass.__module__
        # Passing the name of the decorated class to the new returned class
        # in order to get the test case name and references, i.e. suite_id and case_id
        # in the test results in the console and in the report.
        # Changing __name__ is necessary to make the test name appear in the test results in the console.
        # Changing __qualname__ is necessary to make the test name appear in the test results in the report.
        NewClass.__name__ = DecoratedClass.__name__ + f"-{suite_id}-{case_id}"
        NewClass.__qualname__ = DecoratedClass.__qualname__ + f"-{suite_id}-{case_id}"

        return NewClass

    return generate_modified_class



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Multiprocessing Proxy Auxiliary
*******************************

:module: mp_proxy_auxiliary

:synopsis: concrete implementation of a multiprocessing proxy auxiliary

This auxiliary simply spread all commands and received messages to all
connected auxiliaries. This auxiliary is only usable through mp proxy
connector.

.. code-block:: none

     ___________   ___________         ___________
    |           | |           | ..... |           |
    |   Aux 1   | |   Aux 1   |       |   Aux n   |
    |___________| |___________|       |___________|
          |             |                   |
          |             |                   |
     ___________   ___________         ___________
    |           | |           | ..... |           |
    |Proxy_con 1| |Proxy_con 2|       |Proxy_con n|
    |___________| |___________|       |___________|
          |             |                   |
          |             |                   |
          |             |                   |
     _____________________________________________
    |                                             |
    |               Proxy Auxiliary               |
    |_____________________________________________|
                        |
                        |
     _____________________________________________
    |                                             |
    |               Connector Channel             |
    |_____________________________________________|

.. currentmodule:: mp_proxy_auxiliary

"""
import collections
import logging
import multiprocessing
import sys
import time
from pathlib import Path
from typing import List, Optional, Tuple

from pykiso import AuxiliaryInterface, CChannel, MpAuxiliaryInterface
from pykiso.test_setup.config_registry import ConfigRegistry
from pykiso.test_setup.dynamic_loader import PACKAGE

TraceOptions = collections.namedtuple("TraceOptions", "activate dir name")

log = logging.getLogger(__name__)


class MpProxyAuxiliary(MpAuxiliaryInterface):
    """Proxy auxiliary for multi auxiliaries communication handling.

    ..note :: this auxiliary version is using the multiprocessing
        auxiliary interface.
    """

    def __init__(
        self,
        com: CChannel,
        aux_list: List[str],
        activate_trace: bool = False,
        trace_dir: Optional[str] = None,
        trace_name: Optional[str] = None,
        **kwargs,
    ) -> None:
        """Initialize attributes.

        :param com: Communication connector
        :param aux_list: list of auxiliary's alias
        :param activate_trace: True if the trace is activate otherwise
            False
        :param trace_dir: trace directory path (absolute or relative)
        :param trace_name: trace full name (without file extension)
        """
        self.channel = com
        self.trace_options = TraceOptions(
            activate=activate_trace, dir=trace_dir, name=trace_name
        )
        self.proxy_channels = self.get_proxy_con(aux_list)
        self.logger = None
        self.aux_list = aux_list
        super().__init__(**kwargs)

    def _init_trace(
        self,
        logger,
        activate: bool,
        t_dir: Optional[str] = None,
        t_name: Optional[str] = None,
    ) -> None:
        """Initialize the logging trace for proxy auxiliary received
        message recording and process debugging.

        :param logger: base logger where a FileHandler is added
        :param activate: True if the trace is activate otherwise False
        :param t_dir: trace directory path (absolute or relative)
        :param t_name: trace full name (without file extension)
        """

        if not activate:
            return None

        # Just avoid the case the given trace directory is None
        t_dir = "" if t_dir is None else t_dir
        # if the given log path is not absolute add root path
        # (where pykiso is launched) otherwise take it as it is
        dir_path = (
            (Path() / t_dir).resolve() if not Path(t_dir).is_absolute() else Path(t_dir)
        )
        # if no specific logging file name is given take the default one
        t_name = (
            time.strftime(f"%Y-%m-%d_%H-%M-%S_{t_name}.log")
            if t_name is not None
            else time.strftime("%Y-%m-%d_%H-%M-%S_proxy_logging.log")
        )
        # if path doesn't exists take root path (where pykiso is launched)
        log_path = (
            dir_path / t_name if dir_path.exists() else (Path() / t_name).resolve()
        )

        # configure the file handler and create the tarce file
        log_format = logging.Formatter("%(asctime)s : %(message)s")
        logger.info(f"create proxy trace file at {log_path}")
        handler = logging.FileHandler(log_path, "w+")
        handler.setFormatter(log_format)
        handler.setLevel(logging.DEBUG)
        logger.addHandler(handler)

    def get_proxy_con(self, aux_list: List[str]) -> Tuple[AuxiliaryInterface]:
        """Retrieve all connector associated to all given existing Auxiliaries.

        If auxiliary alias exists but auxiliary instance was not created
        yet, create it immediately using ConfigRegistry _aux_cache.

        :param aux_list: list of auxiliary's alias

        :return: tuple containing all connectors associated to
            all given auxiliaries
        """
        channel_inst = []

        for aux_name in aux_list:
            aux_inst = sys.modules.get(f"{PACKAGE}.auxiliaries.{aux_name}")
            if aux_inst is not None:
                self._check_compatibility(aux_inst)
                channel_inst.append(aux_inst.channel)
            # check if the given aux_name is in the available aux
            # alias list
            elif aux_name in ConfigRegistry.get_auxes_alias():
                log.warning(
                    f"Auxiliary : {aux_name} is not using import magic mechanism (pre-loaded)"
                )
                # load it using ConfigRegistry _aux_cache
                aux_inst = ConfigRegistry._linker._aux_cache.get_instance(aux_name)
                self._check_compatibility(aux_inst)
                channel_inst.append(aux_inst.channel)
            # the given auxiliary alias doesn't exist or refer to a
            # invalid one
            else:
                log.error(f"Auxiliary : {aux_name} doesn't exist")

        return tuple(channel_inst)

    @staticmethod
    def _check_compatibility(aux: AuxiliaryInterface) -> None:
        """Check if the given auxiliary is proxy compatible.

        :param aux: auxiliary instance to check

        :raises NotImplementedError: if is_proxy_capable flag is False
        """
        if not aux.is_proxy_capable:
            raise NotImplementedError(
                f"Auxiliary {aux} is not compatible with a proxy auxiliary"
            )

    def _create_auxiliary_instance(self) -> bool:
        """Open current associated channel.

        :return: if channel creation is successful return True otherwise false
        """
        try:
            self.logger.info("Create auxiliary instance")
            self.logger.info("Enable channel")
            self.channel.open()
            return True
        except Exception:
            self.logger.exception("Error encouting during channel creation")
            self.stop()
            return False

    def _delete_auxiliary_instance(self) -> bool:
        """Close current associated channel.

        :return: always True
        """
        try:
            self.logger.info("Delete auxiliary instance")
            self.channel.close()
            # in very rare case if a queue is populated, ITF could be
            # blocked in a infinite loop. This is a workaround, and not
            # the final implementation.
            self._empty_queue(self.queue_in)
            for conn in self.proxy_channels:
                self._empty_queue(conn.queue_in)
                self._empty_queue(conn.queue_out)
        except Exception:
            self.logger.exception("Error encouting during channel closure")

        return True

    @staticmethod
    def _empty_queue(queue: multiprocessing.Queue) -> None:
        """Empty a given queue by simply calling get until queue  is
        empty.

        :param queue: queue to clear
        """
        while not queue.empty():
            queue.get_nowait()

    def _run_command(self) -> None:
        """Run all commands present in each proxy connectors queue in
        by sending it over current associated CChannel.

        In addition, all commands are dispatch to others auxiliaries
        using proxy connector queue out.
        """
        for conn in self.proxy_channels:
            if not conn.queue_in.empty():
                args, kwargs = conn.queue_in.get()
                message = kwargs.get("msg")
                if message is not None:
                    self._dispatch_command(
                        message=message,
                        remote_id=kwargs.get("remote_id"),
                        con_use=conn,
                    )

                self.channel._cc_send(*args, **kwargs)

    def _dispatch_command(self, message: bytes, con_use: CChannel, remote_id: int):
        """Dispatch the current command to others connected auxiliaries.

        This action is performed by populating the queue out from each
        proxy connectors.

        :param message: message to send
        :param con_use: current proxy connector where the command come from
        :param remote_id: if CAN is used, CAN frame ID on which
            the message will be sent
        """
        for conn in self.proxy_channels:
            if conn != con_use:
                conn.queue_out.put([message, remote_id])

    def _abort_command(self) -> None:
        """Not Used."""
        return True

    def _receive_message(self, timeout_in_s: float = 0) -> None:
        """When no request are sent this method is called by
        AuxiliaryInterface run method. At each message received, this
        method will populate each proxy connectors queue out.

        :param timeout_in_s: maximum amount of time in second to wait
            for a message.
        """
        try:
            received_data, source = self.channel.cc_receive(
                timeout=timeout_in_s, raw=True
            )
            # if data are received, populate connected proxy connectors
            # queue out
            if received_data is not None:
                self.logger.debug(
                    f"raw data : {received_data.hex()} || source : {source} || channel : {self.channel.name}"
                )
                for conn in self.proxy_channels:
                    conn.queue_out.put([received_data, source])
        except Exception:
            self.logger.exception(
                f"encountered error while receiving message via {self.channel}"
            )

    def run(self) -> None:
        """Run function of the auxiliary process."""
        # initialize logging mechanism at process startup
        self.initialize_loggers()
        # initialize trace for message recording
        self._init_trace(self.logger, *self.trace_options)

        while not self.stop_event.is_set():
            # Step 1: Check if a request is available & process it
            request = ""
            # Check if a request was received
            if not self.queue_in.empty():
                request = self.queue_in.get_nowait()
            # Process the request
            if request == "create_auxiliary_instance" and not self.is_instance:
                # Call the internal instance creation method
                self.is_instance = self._create_auxiliary_instance()
                # Enqueue the result for the request caller
                self.queue_out.put(self.is_instance)
            elif request == "delete_auxiliary_instance" and self.is_instance:
                # Call the internal instance delete method
                self.is_instance = not (self._delete_auxiliary_instance())
                # Enqueue the result for the request caller
                self.queue_out.put(self.is_instance)
            elif request != "":
                # A request was received but could not be processed
                self.logger.warning(
                    f"Unknown request '{request}', will not be processed!"
                )
                self.logger.warning(f"Aux status: {self.__dict__}")

            # Step 2: Send stack command and propagate them to others
            # connected auxiliaires and check if something was received
            # if instance was created
            if self.is_instance:
                self._run_command()
                self._receive_message(timeout_in_s=0)

            # Free up cpu usage when auxiliary is suspended
            if not self.is_instance:
                time.sleep(0.050)

        # Thread stop command was set
        self.logger.info("{} was stopped".format(self))
        # Delete auxiliary external instance if not done
        if self.is_instance:
            self._delete_auxiliary_instance()



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Communication Channel via socket
********************************

:module: cc_socket

:synopsis: connector for communication via socket

.. currentmodule:: cc_socket

"""
import logging
import socket
from typing import Union

from pykiso import CChannel

log = logging.getLogger(__name__)


class CCTcpip(CChannel):
    """Connector channel used to communicate via socket"""

    def __init__(self, dest_ip: str, dest_port: int, max_msg_size: int = 256, **kwargs):
        """Initialize channel settings.

        :param dest_ip: destination ip address
        :param dest_port: destination port
        :param max_msg_size: the maximum amount of data to be received
            at once
        """
        super().__init__(**kwargs)
        self.dest_ip = dest_ip
        self.dest_port = int(dest_port)
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.max_msg_size = max_msg_size
        # Set a timeout to send the signal to the GIL to change thread.
        # In case of a multi-threading system, all tasks will be called one after the other.
        self.timeout = 1e-6

    def _cc_open(self) -> None:
        """Connect to socket with configured port and IP address."""
        log.info(
            f"Connection to socket at address {self.dest_ip} port {self.dest_port}"
        )
        self.socket.settimeout(3)
        self.socket.connect((self.dest_ip, self.dest_port))

    def _cc_close(self) -> None:
        """Close UDP socket."""
        log.info(
            f"Disconnect from socket at address {self.dest_ip}, port {self.dest_port}"
        )
        self.socket.close()

    def _cc_send(self, msg: bytes or str, raw: bool = False) -> None:
        """Send a message via socket.

        :param msg: message to send
        :param raw: is the message in a raw format (True) or is it a string (False)?
        """
        if msg is not None and raw is False:
            msg = msg.encode()

        log.debug(f"Sending {msg} via socket to {self.dest_ip}")
        self.socket.send(msg)

    def _cc_receive(self, timeout=0.01, raw: bool = False) -> Union[bytes, str]:
        """Read message from socket.

        :param timeout: time in second to wait for reading a message
        :param raw: should the message be returned raw or should it be interpreted as a
            pykiso.Message?

        :return: Message if successful, otherwise none
        """
        self.socket.settimeout(timeout or self.timeout)

        msg_received = ""
        try:
            msg_received = self.socket.recv(self.max_msg_size)

            if not raw:
                msg_received = msg_received.decode().strip()

            log.debug(f"Socket at {self.dest_ip} received: {msg_received}")
        except socket.timeout:
            log.exception(
                f"encountered timeout error while receiving message via {self}"
            )
        except Exception:
            log.exception(f"encountered error while receiving message via {self}")

        return msg_received



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Device Under Test Auxiliary
***************************

:module: DUTAuxiliary

:synopsis: The Device Under Test auxiliary allow to flash and run test
    on the target using the connector provided.

.. currentmodule:: dut_auxiliary

"""
import logging

from pykiso import AuxiliaryInterface, CChannel, Flasher, message

log = logging.getLogger(__name__)


class DUTAuxiliary(AuxiliaryInterface):
    """Device Under Test(DUT) auxiliary implementation."""

    def __init__(
        self,
        name: str = None,
        com: CChannel = None,
        flash: Flasher = None,
        **kwargs,
    ):
        """Constructor.

        :param name: Alias of the auxiliary instance
        :param com: Communication connector
        :param flash: flash connector
        """
        super().__init__(name=name, **kwargs)
        self.channel = com
        self.flash = flash
        self._is_suspend = False
        log.debug(f"com is {com}")
        log.debug(f"flash is {flash}")

    def _create_auxiliary_instance(self) -> bool:
        """Create auxiliary instance flash and connect.

        :return: True if everything was successful otherwise False
        """

        log.info("Create auxiliary instance")

        # Flash the device if flash connector provided
        if self.flash is not None and not self._is_suspend:
            log.info("Flash target")
            try:
                with self.flash as flasher:
                    flasher.flash()
            except Exception as e:
                # stop the thread if the flashing failed
                log.exception(f"Error occurred during flashing : {e}")
                log.fatal("Stopping the auxiliary")
                self.stop()
                return False  # Prevent to open channels by returning error state

        # Enable the communication through the connector
        log.info("Open channel communication")
        return_code = False
        try:
            self.channel.open()
        except Exception:
            log.exception("Unable to open channel communication")
            self.stop()
        else:
            # Ping-pong test
            return_code = self._send_ping_command(2, 1)

        # Return output
        return return_code

    def _delete_auxiliary_instance(self) -> bool:
        """Close the connector communication.

        :return: always True
        """

        log.info("Close the DUT auxiliary instance")
        self.channel.close()
        return True

    def suspend(self) -> None:
        """Suspend DutAuxiliary's run.

        Set _is_suspend flag to True to avoid a re-flash sequence.
        """
        self._is_suspend = True
        super().suspend()

    def resume(self) -> None:
        """Resume DutAuxiliary's run.

        Set _is_suspend flag to False in order to re-activate flash
        sequence in case of e.g. a futur abort command.
        """
        self._is_suspend = False
        super().resume()

    def _run_command(
        self, cmd_message: message.Message, cmd_data: bytes = None
    ) -> bool:
        """Run a command for the auxiliary.

        :param cmd_message: command in form of a message to run
        :param cmd_data: payload data for the command

        :return: True - Successfully received bv the instance / False - Failed sending
        """

        log.info(f"Send test request: {cmd_message}")
        return self._send_and_wait_ack(cmd_message, 2, 2)

    def _abort_command(self) -> bool:
        """Send an abort command and reset the auxiliary if needed.

        :return: True if ACK is received otherwise False
        """

        log.info("Send abort request")
        # Try a soft abort
        msg = message.Message(
            message.MessageType.COMMAND, message.MessageCommandType.ABORT
        )  # TODO verify if this generation can happened somewhere else
        result = self._send_and_wait_ack(msg, 2, 2)
        # If not successful, do hard reset
        if result is False:
            self._delete_auxiliary_instance()
            self._create_auxiliary_instance()
        return result

    def _receive_message(self, timeout_in_s: float) -> message.Message:
        """Get message from the device under test.

        :param timeout_in_s: Time in ms to wait for one try

        :returns: receive message
        """

        # Read message on the channel
        received_message = self.channel.cc_receive(timeout_in_s)
        if received_message is not None:
            # Send ack
            self.channel._cc_send(
                msg=received_message.generate_ack_message(message.MessageAckType.ACK)
            )
        # Return message
        return received_message

    def _send_ping_command(self, timeout: int, tries: int) -> bool:
        """Ping Pong test to confirm the communication state.

        :param timeout: Time in ms to wait for one try
        :param tries: Number of tries to send the message

        :return: True if the target answer
        """
        number_of_tries = 0
        is_pong_ack = False

        # Empty memory in case target start by sending a message
        self.channel.cc_receive(timeout=1)

        # Try Ping-Pong
        while number_of_tries < tries:
            log.info(f"Ping-Pong try n: {number_of_tries}")

            # Increase number of tries
            number_of_tries += 1

            # Send a ping message
            ping_request = message.Message(test_suite=0, test_case=0)
            self.channel.cc_send(msg=ping_request)

            # Receive the message
            pong_response = self.channel.cc_receive(timeout)

            # Validate ping pong
            log.debug(f"ping: {ping_request}")
            log.debug(f"pong: {pong_response}")
            # testing if is tuple instance to avoid infinite loop when
            # DUT auxiliary is bounded with a proxy auxiliary
            if pong_response is not None and not isinstance(pong_response, tuple):
                if ping_request.check_if_ack_message_is_matching(pong_response):
                    log.info("Ping-Pong succeed")
                    is_pong_ack = True
                    break
                else:
                    log.warning(
                        f"Received {pong_response} not matching {ping_request}!"
                    )
                    continue  # A NACK got received # TODO later on we should have "not supported" and ignore it than.

        return is_pong_ack

    def _send_and_wait_ack(
        self, message_to_send: message.Message, timeout: int, tries: int
    ) -> bool:
        """Send via the channel a message and wait for an acknowledgement.

        :param message_to_send: Message you want to send out
        :param timeout: Time in ms to wait for one try
        :param tries: Number of tries to send the message

        :returns: True - Ack, False - Nack
        """

        number_of_tries = 0
        result = False

        while number_of_tries < tries:
            log.info(f"Run send try n: {number_of_tries}")

            # Increase number of tries
            number_of_tries += 1

            # Send the message
            self.channel.cc_send(msg=message_to_send)

            # Wait until we get something back
            received_message = self.channel.cc_receive(timeout)

            # Check the outcome
            if received_message is None:
                continue  # Next try will occur
            else:
                log.debug(f"message_to_send  {message_to_send}")
                log.debug(f"received_message {received_message}")
                if message_to_send.check_if_ack_message_is_matching(received_message):
                    log.info(f"{message_to_send} sent")
                    result = True
                    break
                else:
                    log.warning(
                        f"Received {received_message} not matching with {message_to_send}!"
                    )
                    continue  # A NACK got received # TODO later on we should have "not supported" and ignore it than.
        return result



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Proxy Channel
*************

:module: cc_proxy

:synopsis: CChannel implementation for multi-auxiliary usage.

CCProxy channel was created, in order to enable the connection of
multiple auxiliaries on one and only one CChannel. This CChannel
has to be used with a so called proxy auxiliary.

.. currentmodule:: cc_proxy

"""

import logging
import queue
from multiprocessing import Queue
from typing import Tuple, Union

from pykiso import Message
from pykiso.connector import CChannel

ProxyReturn = Union[
    Tuple[bytes, int], Tuple[bytes, None], Tuple[Message, None], Tuple[None, None]
]

log = logging.getLogger(__name__)


class CCProxy(CChannel):
    """Proxy CChannel for multi auxiliary usage."""

    def __init__(self, **kwargs):
        """Initialize attributes."""
        super().__init__(**kwargs)
        self.queue_in = None
        self.queue_out = None
        self.timeout = 1

    def _cc_open(self) -> None:
        """Open proxy channel."""
        self.queue_in = Queue()
        self.queue_out = Queue()
        log.debug("Open proxy channel")

    def _cc_close(self) -> None:
        """Close proxy channel."""
        self.queue_in = Queue()
        self.queue_out = Queue()
        log.debug("Close proxy channel")

    def _cc_send(self, *args: tuple, **kwargs: dict) -> None:
        """Populate the queue in of the proxy connector.

        :param args: tuple containing positionnal arguments
        :param kwargs: dictionary containing named arguments
        """
        log.debug(f"put at proxy level: {args} {kwargs}")
        self.queue_in.put((args, kwargs))

    def _cc_receive(self, timeout: float = 0.1, raw: bool = False) -> ProxyReturn:
        """Depopulate the queue out of the proxy connector.

        :param timeout: not used
        :param raw: not used

        :return: raw bytes and source when it exist. if queue timeout
            is reached return None
        """

        try:
            raw_msg, source = self.queue_out.get(True, self.timeout)
            log.debug(f"received at proxy level : {raw_msg} || source {source}")
            return raw_msg, source
        except queue.Empty:
            return None, None



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

#!/usr/bin/env python
# -*- encoding: utf-8 -*-

import io
from glob import glob
from os.path import basename, dirname, join, splitext

from setuptools import find_packages, setup


def read(*names, **kwargs):
    with io.open(
        join(dirname(__file__), *names), encoding=kwargs.get("encoding", "utf8")
    ) as fh:
        return fh.read()


install_requires = [
    "pyserial",
    "click",
    "pyyaml",
    "pylink-square",
    "python-can[pcan]",
    "unittest-xml-reporting",
    "robotframework==3.2.2",
    "pyvisa",
    "pyvisa-py",
]

setup(
    name="pykiso",
    version="0.16.0",
    license="Eclipse Public License - v 2.0",
    description="Embedded integration testing framework.",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    author="Sebastian Fischer",
    author_email="sebastian.fischer@de.bosch.com",
    url="https://github.com/eclipse/kiso-testing",
    packages=find_packages("src"),
    package_dir={"": "src"},
    py_modules=[splitext(basename(path))[0] for path in glob("src/*.py")],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Eclipse Public License 2.0 (EPL-2.0)",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Utilities",
    ],
    keywords=[
        "testing",
        "integration testing",
        "framework",
        "testing framework",
    ],
    python_requires=">=3.6",
    install_requires=install_requires,
    tests_require=["pytest", "pytest-mock", "coverage"],
    extras_require={},
    setup_requires=[],
    entry_points={
        "console_scripts": [
            "pykiso = pykiso.cli:main",
            "instrument-control = pykiso.lib.auxiliaries.instrument_control_auxiliary.instrument_control_cli:main",
        ]
    },
)



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
test_xml_result
***************

:module: test_xml_result

:synopsis: overwrite xmlrunner.result to be able to add additional data into the xml report.

.. currentmodule:: test_xml_result

"""

import copy
import json
import sys
from io import TextIOWrapper
from xml.dom.minidom import Document, Element

import xmlrunner.result
import xmlrunner.runner


class TestInfo(xmlrunner.result._TestInfo):
    """This class keeps useful information about the execution of a test method.
    Used by XmlTestResult
    """

    def __init__(
        self,
        test_result: xmlrunner.runner._XMLTestResult,
        test_method,
        outcome: int = 0,
        err=None,
        subTest=None,
        filename: str = None,
        lineno: bool = None,
        doc: str = None,
    ):
        """Initialize the TestInfo class and append additional tag
           that have to be stored for each test

        :param test_result: test result class
        :param test_method: test method (dynamically created eg: test_case.MyTest2-1-2)
        :param outcome: result of the test (SUCCESS, FAILURE, ERROR, SKIP)
        :param err: error cached during test
        :param subTest: optional, refer the test id and the test description
        :param filename: name of the file
        :param lineno: store the test line number
        :param doc: additional documentation to store
        """
        super().__init__(
            test_result, test_method, outcome, err, subTest, filename, lineno, doc
        )

        # store extra tag
        self.test_ids = json.dumps(test_method.test_ids)


class XmlTestResult(xmlrunner.runner._XMLTestResult):
    """Test result class that can express test results in a XML report.
    Used by XMLTestRunner
    """

    def __init__(
        self,
        stream: TextIOWrapper = sys.stderr,
        descriptions: bool = True,
        verbosity: int = 1,
        elapsed_times: bool = True,
        properties=None,
        infoclass: xmlrunner.result._TestInfo = TestInfo,
    ):
        """Initialize the _XMLTestResult class.

        :param stream: buffered text interface to a buffered raw stream
        :param descriptions: include description of the test
        :param verbosity: print output into the console
        :param elapsed_times: include the time spend on the test
        :param properties: junit testsuite properties
        :param infoclass: class containing the test information
        """
        super().__init__(
            stream=stream,
            descriptions=descriptions,
            verbosity=verbosity,
            elapsed_times=elapsed_times,
            properties=properties,
            infoclass=infoclass,
        )

    # save the original method that will be overwritten
    report_testcase = copy.deepcopy(xmlrunner.runner._XMLTestResult._report_testcase)

    @staticmethod
    def _report_testcase(
        test_result: TestInfo, xml_testsuite: Element, xml_document: Document
    ):
        """Callback function that create the test case section into the XML document.

        :param test_result: result of a test run
        :param xml_testsuite: xml test suite to be written
        :param xml_document: xml document base
        """
        # call the original method
        XmlTestResult.report_testcase(test_result, xml_testsuite, xml_document)

        # here can be added additional tags that have to be stored into the xml test report
        xml_testsuite.setAttribute("test_ids", str(test_result.test_ids))

    # Catch and redefine xmlrunner.runner._XMLTestResult._report_testcase
    xmlrunner.runner._XMLTestResult._report_testcase = _report_testcase



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Virtual Communication Channel for tests
***************************************

:module: cc_example


.. warning: Still under test
"""

import logging
from typing import Optional, Union

from pykiso import connector, message

log = logging.getLogger(__name__)


class CCExample(connector.CChannel):
    """Only use for development purpose.

    This channel simply handle basic TestApp response mechanism.
    """

    def __init__(self, name: Optional[str] = None, **kwargs):
        """Initialize attributes.

        :param name: name of the communication channel
        """
        super().__init__(name=name)
        self.options = kwargs
        self.last_received_message = None
        self.report_requested_message = None

    def _cc_open(self) -> None:
        """Open the channel."""
        log.info("open channel")

    def _cc_close(self) -> None:
        """Close the channel."""
        log.info("close channel")

    def _cc_send(self, msg: message.Message, raw: bool = False) -> None:
        """Sends the message on the channel.

        :param msg: message to send, should be Message type like.
        :param raw: if raw is false serialize it using Message serialize.

        :raise NotImplementedError: sending raw bytes is not supported.
        """
        if raw:
            raise NotImplementedError()
        log.debug("Send: {}".format(msg))
        # Exit if ack sent
        if msg.get_message_type() == message.MessageType.ACK:
            return

        # Else save received message
        self.last_received_message = msg.serialize()

        # Check if it is a run test-case message and save it for the report
        if (
            msg.sub_type in set(message.MessageCommandType)
            and msg.sub_type != message.MessageCommandType.PING
            and msg.sub_type != message.MessageCommandType.ABORT
        ):

            self.report_requested_message = msg.serialize()

    def _cc_receive(
        self, timeout: float = 0.1, raw: bool = False
    ) -> Union[message.Message, None]:
        """Reads from the channel - decorator usage for test.

        :param timeout: not use
        :param raw: if raw is false serialize it using Message serialize.

        :raise NotImplementedError: receiving raw bytes is not supported.

        :return: Message if successful, otherwise None
        """
        if raw:
            raise NotImplementedError()
        if self.last_received_message is not None:
            # Transform into ack
            r_message = message.Message.parse_packet(self.last_received_message)
            r_message.msg_type = message.MessageType.ACK
            r_message.sub_type = message.MessageAckType.ACK
            # Delete the stored raw message
            self.last_received_message = None
            # Return the ACK
            log.debug("Receive: {}".format(r_message))
            return r_message
        elif self.report_requested_message is not None:
            # Transform message to ACK
            r_message = message.Message.parse_packet(self.report_requested_message)
            r_message.msg_type = message.MessageType.REPORT
            r_message.sub_type = message.MessageReportType.TEST_PASS
            # Delete the stored raw message
            self.report_requested_message = None
            # Return REPORT
            log.debug("Receive: {}".format(r_message))
            return r_message



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Integration Test Framework
**************************

:module: cli

:synopsis: Entry point to the integration test framework.

.. currentmodule:: cli


"""
import collections
import logging
import pprint
import sys
import time
from pathlib import Path
from typing import NamedTuple, Optional

import click

from . import __version__
from .config_parser import parse_config
from .global_config import Grabber
from .test_coordinator import test_execution
from .test_setup.config_registry import ConfigRegistry
from .types import PathType

LogOptions = collections.namedtuple("LogOptions", "log_path log_level report_type")

# use to store the selected logging options
log_options: Optional[NamedTuple] = None


def initialize_logging(
    log_path: PathType, log_level: str, report_type: str = None
) -> logging.Logger:
    """Initialize the logging.

    Sets the general log level, output file or STDOUT and the
    logging format.

    :param log_path: path to the logfile
    :param log_level: any of DEBUG, INFO, WARNING, ERROR
    :param report_type: expected report type (junit, text,...)

    :returns: configured Logger
    """
    root_logger = logging.getLogger()
    log_format = logging.Formatter(
        "%(asctime)s [%(levelname)s] %(module)s:%(lineno)d: %(message)s"
    )
    levels = {
        "DEBUG": logging.DEBUG,
        "INFO": logging.INFO,
        "WARNING": logging.WARNING,
        "ERROR": logging.ERROR,
    }

    # update logging options
    global log_options
    log_options = LogOptions(log_path, log_level, report_type)

    # if log_path is given create use a logging file handler
    if log_path is not None:
        log_path = Path(log_path)
        if log_path.is_dir():
            fname = time.strftime("%Y-%m-%d_%H-%M-test.log")
            log_path = log_path / fname
        file_handler = logging.FileHandler(log_path, "w+")
        file_handler.setFormatter(log_format)
        file_handler.setLevel(levels[log_level])
        root_logger.addHandler(file_handler)
    # if log_path is not given and report type is not junit just
    # instanciate a logging StreamHandler
    if log_path is None and report_type != "junit":
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(log_format)
        stream_handler.setLevel(levels[log_level])
        root_logger.addHandler(stream_handler)
    # if report_type is junit use sys.stdout as stream
    if report_type == "junit":
        # flush all StreamHandler
        for handler in root_logger.handlers:
            if isinstance(handler, logging.StreamHandler):
                handler.flush()
        # but keep FileHandler
        root_logger.handlers = [
            handler
            for handler in root_logger.handlers
            if isinstance(handler, logging.FileHandler)
        ]
        stream_handler = logging.StreamHandler(sys.stdout)
        stream_handler.setFormatter(log_format)
        stream_handler.setLevel(levels[log_level])
        root_logger.addHandler(stream_handler)

    root_logger.setLevel(levels[log_level])

    return logging.getLogger(__name__)


def get_logging_options() -> LogOptions:
    """Simply return the previous logging options.

    :return: logging options log path, log level and report type
    """
    return log_options


@click.command(context_settings={"help_option_names": ["-h", "--help"]})
@click.option(
    "-c",
    "--test-configuration-file",
    required=True,
    type=click.Path(exists=True, dir_okay=False, readable=True),
    help="path to the test configuration file (in YAML format)",
)
@click.option(
    "-l",
    "--log-path",
    required=False,
    default=None,
    type=click.Path(writable=True),
    help="path to log-file or folder. If not set will log to STDOUT",
)
@click.option(
    "--log-level",
    required=False,
    default="INFO",
    type=click.Choice(
        "DEBUG INFO WARNING ERROR".split(" "),
        case_sensitive=False,
    ),
    help="set the verbosity of the logging",
)
@click.option(
    "--junit",
    "report_type",
    flag_value="junit",
    required=False,
    help="enables the generation of a junit report",
)
@click.option(
    "--text",
    "report_type",
    flag_value="text",
    required=False,
    default=True,
    help="default, test results are only displayed in the console",
)
@click.option(
    "--variant",
    multiple=True,
    type=str,
    default=None,
    help="allow the user to execute a subset of tests based on variants",
)
@click.option(
    "--branch-level",
    multiple=True,
    type=str,
    default=None,
    help="allow the user to execute a subset of tests based on branch levels",
)
@click.option(
    "--failfast",
    is_flag=True,
    help="stop the test run on the first error or failure",
)
@click.argument("pattern", required=False)
@click.version_option(__version__)
@Grabber.grab_cli_config
def main(
    test_configuration_file: PathType,
    log_path: PathType = None,
    log_level: str = "INFO",
    report_type: str = "text",
    variant: Optional[tuple] = None,
    branch_level: Optional[tuple] = None,
    pattern: Optional[str] = None,
    failfast: bool = False,
):
    """Embedded Integration Test Framework - CLI Entry Point.

    PATTERN: overwrite the test filter pattern from the YAML file (optional)

    \f
    :param test_configuration_file: path to the YAML config file
    :param log_path: path to directory or file to write logs to
    :param log_level: any of DEBUG, INFO, WARNING, ERROR
    :param report_type: if "test", the standard report, if "junit", a junit report is generated
    :param variant: allow the user to execute a subset of tests based on variants
    :param branch_level: allow the user to execute a subset of tests based on branch levels
    :param pattern: overwrite the pattern from the YAML file for easier testdevelopment
    :param failfast: stop the test run on the first error or failure
    """
    # Set the logging
    logger = initialize_logging(log_path, log_level, report_type)
    # Get YAML configuration
    cfg_dict = parse_config(test_configuration_file)
    # Run tests
    logger.debug("cfg_dict:\n{}".format(pprint.pformat(cfg_dict)))

    ConfigRegistry.register_aux_con(cfg_dict)

    exit_code = test_execution.execute(
        cfg_dict, report_type, variant, branch_level, pattern, failfast
    )
    ConfigRegistry.delete_aux_con()
    sys.exit(exit_code)
